# coding: utf-8
"""
Logic module for management of instruments

Instruments are used to group actions together and provide instrument
responsible users with permissions for samples or measurements produced
using the instrument.

Similar to actions, instruments cannot be deleted. However, all information
about an instrument may be altered.
"""

import dataclasses
import typing

from .components import check_component_exists
from .. import db
from .. import models
from ..models.instruments import instrument_user_association_table
from . import users, errors, components, locations, objects, topics
from .utils import cache


@dataclasses.dataclass(frozen=True)
class Instrument:
    """
    This class provides an immutable wrapper around models.instruments.Instrument.
    """
    id: int
    responsible_users: typing.List[users.User]
    users_can_create_log_entries: bool
    users_can_view_log_entries: bool
    create_log_entry_default: bool
    is_hidden: bool
    name: typing.Dict[str, str]
    notes: typing.Dict[str, str]
    notes_is_markdown: bool
    description: typing.Dict[str, str]
    description_is_markdown: bool
    short_description: typing.Dict[str, str]
    short_description_is_markdown: bool
    fed_id: typing.Optional[int]
    component_id: typing.Optional[int]
    component: typing.Optional[components.Component]
    location_id: typing.Optional[int]
    location: typing.Optional[locations.Location]
    object_id: typing.Optional[int]
    show_linked_object_data: bool
    topics: typing.List[topics.Topic]

    @classmethod
    def from_database(cls, instrument: models.Instrument) -> 'Instrument':
        return Instrument(
            id=instrument.id,
            responsible_users=[
                users.User.from_database(user)
                for user in instrument.responsible_users
            ],
            users_can_create_log_entries=instrument.users_can_create_log_entries,
            users_can_view_log_entries=instrument.users_can_view_log_entries,
            create_log_entry_default=instrument.create_log_entry_default,
            is_hidden=instrument.is_hidden,
            name=instrument.name,
            notes=instrument.notes,
            notes_is_markdown=bool(instrument.notes_is_markdown),
            description=instrument.description,
            description_is_markdown=bool(instrument.description_is_markdown),
            short_description=instrument.short_description,
            short_description_is_markdown=bool(instrument.short_description_is_markdown),
            fed_id=instrument.fed_id,
            component_id=instrument.component_id,
            component=components.Component.from_database(instrument.component) if instrument.component is not None else None,
            location_id=instrument.location_id,
            location=locations.Location.from_database(instrument.location) if instrument.location is not None else None,
            object_id=instrument.object_id,
            show_linked_object_data=instrument.show_linked_object_data,
            topics=[topics.Topic.from_database(topic) for topic in instrument.topics],
        )

    def __repr__(self) -> str:
        return f"<{type(self).__name__}(id={self.id!r})>"


@typing.overload
def create_instrument(
        *,
        description_is_markdown: bool = False,
        users_can_create_log_entries: bool = False,
        users_can_view_log_entries: bool = False,
        notes_is_markdown: bool = False,
        create_log_entry_default: bool = False,
        is_hidden: bool = False,
        short_description_is_markdown: bool = False,
        fed_id: None = None,
        component_id: None = None,
        show_linked_object_data: bool = True,
) -> Instrument:
    ...


@typing.overload
def create_instrument(
        *,
        description_is_markdown: bool = False,
        users_can_create_log_entries: bool = False,
        users_can_view_log_entries: bool = False,
        notes_is_markdown: bool = False,
        create_log_entry_default: bool = False,
        is_hidden: bool = False,
        short_description_is_markdown: bool = False,
        fed_id: int,
        component_id: int,
        show_linked_object_data: bool = True,
) -> Instrument:
    ...


def create_instrument(
        *,
        description_is_markdown: bool = False,
        users_can_create_log_entries: bool = False,
        users_can_view_log_entries: bool = False,
        notes_is_markdown: bool = False,
        create_log_entry_default: bool = False,
        is_hidden: bool = False,
        short_description_is_markdown: bool = False,
        fed_id: typing.Optional[int] = None,
        component_id: typing.Optional[int] = None,
        show_linked_object_data: bool = True,
) -> Instrument:
    """
    Creates a new instrument.

    :param description_is_markdown: whether the description contains Markdown
    :param users_can_create_log_entries: whether users can create log
        entries for this instrument
    :param users_can_view_log_entries: whether users can view the log
        entries for this instrument
    :param notes_is_markdown: whether the notes contain Markdown
    :param create_log_entry_default: the default for whether a log
        entry should be created during object creation by instrument scientists
    :param is_hidden: whether this instrument is hidden
    :param short_description_is_markdown: whether the short description
        contains Markdown
    :param fed_id: the ID of the related instrument at the exporting component
    :param component_id: the ID of the exporting component
    :param show_linked_object_data: whether the data for a linked object
        should be shown
    :return: the new instrument
    """
    assert (component_id is None) == (fed_id is None)

    if component_id is not None:
        check_component_exists(component_id)

    instrument = models.Instrument(
        description_is_markdown=description_is_markdown,
        users_can_create_log_entries=users_can_create_log_entries,
        users_can_view_log_entries=users_can_view_log_entries,
        notes_is_markdown=notes_is_markdown,
        create_log_entry_default=create_log_entry_default,
        is_hidden=is_hidden,
        short_description_is_markdown=short_description_is_markdown,
        fed_id=fed_id,
        component_id=component_id,
        show_linked_object_data=show_linked_object_data,
    )
    db.session.add(instrument)
    db.session.commit()
    return Instrument.from_database(instrument)


def get_instruments() -> typing.List[Instrument]:
    """
    Returns all instruments.

    :return: the list of instruments
    """
    return [
        Instrument.from_database(instrument)
        for instrument in models.Instrument.query.all()
    ]


@cache
def check_instrument_exists(
        instrument_id: int
) -> None:
    """
    Check whether an instrument with the given instrument ID exists.

    :param instrument_id: the ID of an existing instrument
    :raise errors.InstrumentDoesNotExistError: when no instrument with the given
        instrument ID exists
    """
    if not db.session.query(db.exists().where(models.Instrument.id == instrument_id)).scalar():
        raise errors.InstrumentDoesNotExistError()


def get_mutable_instrument(
        instrument_id: int,
        component_id: typing.Optional[int] = None
) -> models.Instrument:
    """
    Get the mutable instrument instance to perform changes in the database on.

    :param instrument_id: the ID of an existing instrument
    :param component_id: the ID of an existing component, or None
    :return: the instrument
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    """
    instrument: typing.Optional[models.Instrument]
    if component_id is None:
        instrument = models.Instrument.query.filter_by(id=instrument_id).first()
    else:
        instrument = models.Instrument.query.filter_by(fed_id=instrument_id, component_id=component_id).first()
    if instrument is None:
        if component_id is not None:
            check_component_exists(component_id)
        raise errors.InstrumentDoesNotExistError()
    return instrument


def get_instrument(
        instrument_id: int,
        component_id: typing.Optional[int] = None
) -> Instrument:
    """
    Return the instrument with the given instrument ID.

    :param instrument_id: the ID of an existing instrument
    :param component_id: the ID of an existing component, or None
    :return: the instrument
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    """
    return Instrument.from_database(get_mutable_instrument(instrument_id, component_id))


def update_instrument(
        *,
        instrument_id: int,
        description_is_markdown: typing.Optional[bool] = None,
        users_can_create_log_entries: typing.Optional[bool] = None,
        users_can_view_log_entries: typing.Optional[bool] = None,
        notes_is_markdown: typing.Optional[bool] = None,
        create_log_entry_default: typing.Optional[bool] = None,
        is_hidden: typing.Optional[bool] = None,
        short_description_is_markdown: typing.Optional[bool] = None,
        show_linked_object_data: typing.Optional[bool] = None
) -> None:
    """
    Updates the instrument.

    :param instrument_id: the ID of an existing instrument
    :param description_is_markdown: whether the description contains Markdown,
        or None
    :param users_can_create_log_entries: whether or not users can create log
        entries for this instrument, or None
    :param users_can_view_log_entries: whether or not users can view the log
        entries for this instrument, or None
    :param notes_is_markdown: whether the notes contain Markdown, or None
    :param create_log_entry_default: the default for whether or not a log
        entry should be created during object creation by instrument
        scientists, or None
    :param is_hidden: whether or not this instrument is hidden, or None
    :param show_linked_object_data: whether the data for a linked object
        should be shown, or None
    :param short_description_is_markdown: whether the short description
        contains Markdown, or None
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    """
    instrument = models.Instrument.query.filter_by(id=instrument_id).first()
    if instrument is None:
        raise errors.InstrumentDoesNotExistError()
    if description_is_markdown is not None:
        instrument.description_is_markdown = description_is_markdown
    if users_can_create_log_entries is not None:
        instrument.users_can_create_log_entries = users_can_create_log_entries
    if users_can_view_log_entries is not None:
        instrument.users_can_view_log_entries = users_can_view_log_entries
    if notes_is_markdown is not None:
        instrument.notes_is_markdown = notes_is_markdown
    if create_log_entry_default is not None:
        instrument.create_log_entry_default = create_log_entry_default
    if short_description_is_markdown is not None:
        instrument.short_description_is_markdown = short_description_is_markdown
    if is_hidden is not None:
        instrument.is_hidden = is_hidden
    if show_linked_object_data is not None:
        instrument.show_linked_object_data = show_linked_object_data
    db.session.add(instrument)
    db.session.commit()


def add_instrument_responsible_user(instrument_id: int, user_id: int) -> None:
    """
    Adds the user with the given user ID to the list of instrument responsible
    users.

    :param instrument_id: the ID of an existing instrument
    :param user_id: the ID of an existing user
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    :raise errors.UserAlreadyResponsibleForInstrumentError: when the user is
        already responsible for the instrument
    """
    instrument = models.Instrument.query.filter_by(id=instrument_id).first()
    if instrument is None:
        raise errors.InstrumentDoesNotExistError()
    user = users.get_mutable_user(user_id)
    if user in instrument.responsible_users:
        raise errors.UserAlreadyResponsibleForInstrumentError()
    instrument.responsible_users.append(user)
    db.session.add(instrument)
    db.session.commit()


def remove_instrument_responsible_user(instrument_id: int, user_id: int) -> None:
    """
    Removes the user with the given user ID from the list of instrument
    responsible users.

    :param instrument_id: the ID of an existing instrument
    :param user_id: the ID of an existing user
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    :raise errors.UserNotResponsibleForInstrumentError: when the user is not
        responsible for the instrument
    """
    instrument = models.Instrument.query.filter_by(id=instrument_id).first()
    if instrument is None:
        raise errors.InstrumentDoesNotExistError()
    user = users.get_mutable_user(user_id)
    if user not in instrument.responsible_users:
        raise errors.UserNotResponsibleForInstrumentError()
    instrument.responsible_users.remove(user)
    db.session.add(instrument)
    db.session.commit()


def set_instrument_responsible_users(instrument_id: int, user_ids: typing.List[int]) -> None:
    """
    Set the list of instrument responsible users for a an instrument.

    :param instrument_id: the ID of an existing instrument
    :param user_ids: the IDs of existing users
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    :raise errors.UserDoesNotExistError: when no user with one of the given
        user IDs exists
    """
    instrument = models.Instrument.query.filter_by(id=instrument_id).first()
    if instrument is None:
        raise errors.InstrumentDoesNotExistError()
    instrument.responsible_users.clear()
    for user_id in user_ids:
        user = users.get_mutable_user(user_id)
        instrument.responsible_users.append(user)
    db.session.add(instrument)
    db.session.commit()


def get_user_instruments(user_id: int, exclude_hidden: bool = False) -> typing.List[int]:
    """
    Get a list of instruments a user with a given user ID is responsible for.

    :param user_id: the ID of an existing user
    :param exclude_hidden: whether hidden instruments should be excluded
    :return: a list of instrument IDs
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    """
    # ensure that the user exists
    users.check_user_exists(user_id)
    instrument_id_query = db.session.query(
        instrument_user_association_table.c.instrument_id
    ).filter(instrument_user_association_table.c.user_id == user_id)
    if exclude_hidden:
        instrument_id_query = instrument_id_query.join(
            models.Instrument,
            instrument_user_association_table.c.instrument_id == models.Instrument.id
        ).filter(models.Instrument.is_hidden == db.false())
    instrument_ids = [
        instrument_user_association[0]
        for instrument_user_association in instrument_id_query.order_by(instrument_user_association_table.c.instrument_id).all()
    ]
    return instrument_ids


def set_instrument_location(instrument_id: int, location_id: typing.Optional[int]) -> None:
    """
    Set the location of an instrument.

    :param instrument_id: the ID of an existing instrument
    :param location_id: the ID of an existing location
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    :raise errors.LocationDoesNotExistError: when no location with the
        given location ID exists
    """
    if location_id is not None:
        # ensure the location exists
        locations.check_location_exists(location_id)
    instrument = models.Instrument.query.filter_by(id=instrument_id).first()
    if instrument is None:
        raise errors.InstrumentDoesNotExistError()
    instrument.location_id = location_id
    db.session.add(instrument)
    db.session.commit()


def set_instrument_object(instrument_id: int, object_id: typing.Optional[int]) -> None:
    """
    Set the object containing the metadata for an instrument.

    :param instrument_id: the ID of an existing instrument
    :param object_id: the ID of an existing object, or None
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    :raise errors.ObjectDoesNotExistError: when no object with the
        given object ID exists
    :raise errors.InstrumentObjectLinkAlreadyExistsError: if the object is
        already linked to an instrument or the instrument is already linked to
        an object
    """
    if object_id is not None:
        # ensure the object exists
        objects.check_object_exists(object_id)
    instrument = models.Instrument.query.filter_by(id=instrument_id).first()
    if instrument is None:
        raise errors.InstrumentDoesNotExistError()
    if object_id is not None:
        if instrument.object_id is not None:
            raise errors.InstrumentObjectLinkAlreadyExistsError()
        linked_instrument = models.Instrument.query.filter_by(object_id=object_id).first()
        if linked_instrument is not None:
            raise errors.InstrumentObjectLinkAlreadyExistsError()
    instrument.object_id = object_id
    db.session.add(instrument)
    db.session.commit()


def get_instrument_object_links() -> typing.Sequence[typing.Tuple[int, int]]:
    """
    Get the list of linked instrument and object ID pairs.

    :return: a list of tuples containing the instrument ID and object ID of
        each linked instrument and object pair
    """
    instruments = models.Instrument.query.filter(models.Instrument.object_id != db.null()).all()
    return [
        (instrument.id, typing.cast(int, instrument.object_id))
        for instrument in instruments
    ]


def get_instruments_for_topic(
        topic_id: int
) -> typing.List[Instrument]:
    """
    Get the list of instruments assigned to a given topic.

    :param topic_id: the ID of an existing topic
    :return: the list of instruments
    :raise errors.TopicDoesNotExistError: if the topic does not exist
    """
    instruments = models.instruments.Instrument.query.filter(models.Instrument.topics.any(models.topics.Topic.id == topic_id)).all()
    if not instruments:
        topics.check_topic_exists(topic_id)
    return [
        Instrument.from_database(instrument)
        for instrument in instruments
    ]



# coding: utf-8
"""

"""

from http import HTTPStatus
import flask
import flask_login
from flask_babel import _
from flask_wtf import FlaskForm
from wtforms import StringField, PasswordField
from wtforms.validators import Email, Length, InputRequired, ValidationError, EqualTo

from .. import frontend
from ...logic import users, authentication
from ...utils import FlaskResponseT
from ...models import UserType


class CreateOtherUserForm(FlaskForm):
    name = StringField(validators=[InputRequired()])
    email = StringField(validators=[Email()])
    password = PasswordField(validators=[Length(min=3)])
    password_confirmation = PasswordField(validators=[EqualTo('password', message='Please enter the same password as above.')])

    def validate_name(form, field: StringField) -> None:
        if len(field.data) < 1:
            raise ValidationError(_('Please enter a valid user name.'))
        if not all(c in 'abcdefghijklmnopqrstuvwxyz0123456789_' for c in field.data):
            raise ValidationError(_('The name may only contain lower case characters, digits and underscores.'))
        if not authentication.is_login_available(field.data):
            raise ValidationError(_('This name is already being used.'))

    def validate_password(self, field: PasswordField) -> None:
        if field.data and len(field.data.encode('utf-8')) > authentication.MAX_BCRYPT_PASSWORD_LENGTH:
            raise ValidationError(_('The password must be at most %(max_bcrypt_password_length)s bytes long.', max_bcrypt_password_length=authentication.MAX_BCRYPT_PASSWORD_LENGTH))


@frontend.route('/users/create_other_user', methods=['GET', 'POST'])
@flask_login.login_required
def create_other_user() -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(HTTPStatus.UNAUTHORIZED)
    create_other_user_form = CreateOtherUserForm()
    if create_other_user_form.validate_on_submit():
        user = users.create_user(
            name=create_other_user_form.name.data,
            email=create_other_user_form.email.data,
            type=UserType.OTHER
        )
        authentication.add_other_authentication(
            user_id=user.id,
            name=create_other_user_form.name.data,
            password=create_other_user_form.password.data
        )
        flask.flash(_('The user has been created successfully.'), 'success')
        return flask.redirect(flask.url_for('.user_profile', user_id=user.id))
    return flask.render_template(
        'create_other_user.html',
        create_other_user_form=create_other_user_form
    )



import binascii
import copy
import dataclasses
import datetime
import hashlib
import io
import itertools
import json
import os.path
import string
import typing
import urllib.parse
import zipfile
from urllib.parse import urlsplit

import flask
import requests
from flask_babel import gettext
import minisign

from .. import db
from .utils import parse_url, relative_url_for
from .users import User, create_user, set_user_hidden
from .objects import create_eln_import_object, update_object, get_object
from .actions import create_action
from .action_types import ActionType
from .action_permissions import set_action_permissions_for_all_users
from .action_translations import set_action_translation
from .files import create_url_file, create_database_file, update_file_information
from .comments import create_comment
from .schemas.validate_schema import validate_schema
from .schemas.validate import validate
from . import errors, object_log
from .units import get_unit_for_un_cefact_code, ureg
from ..models import eln_imports, users, Language, Permissions, UserType, Object


@dataclasses.dataclass
class ELNImport:
    id: int
    file_name: str
    user_id: int
    upload_utc_datetime: datetime.datetime
    import_utc_datetime: typing.Optional[datetime.datetime]
    invalid_reason: typing.Optional[str]
    user: User
    signed_by: typing.Optional[str]

    @classmethod
    def from_database(cls, eln_import: eln_imports.ELNImport) -> 'ELNImport':
        return cls(
            id=eln_import.id,
            file_name=eln_import.file_name,
            user_id=eln_import.user_id,
            upload_utc_datetime=eln_import.upload_utc_datetime,
            import_utc_datetime=eln_import.import_utc_datetime,
            invalid_reason=eln_import.invalid_reason,
            user=User.from_database(eln_import.user),
            signed_by=eln_import.signed_by
        )

    @property
    def is_valid(self) -> bool:
        return self.invalid_reason is None

    @property
    def imported(self) -> bool:
        return self.import_utc_datetime is not None


@dataclasses.dataclass(frozen=True)
class ParsedELNComment:
    text: str
    date_created: datetime.datetime
    author_id: typing.Optional[str]


@dataclasses.dataclass(frozen=True)
class ParsedELNVersion:
    version_id: int
    data: typing.Optional[typing.Dict[str, typing.Any]]
    schema: typing.Optional[typing.Dict[str, typing.Any]]
    user_id: typing.Optional[str]
    date_created: typing.Optional[datetime.datetime]


@dataclasses.dataclass(frozen=True)
class ParsedELNDatabaseFile:
    name: str
    title: typing.Optional[str]
    description: typing.Optional[str]
    content: bytes
    user_id: typing.Optional[str]
    date_created: typing.Optional[datetime.datetime]


@dataclasses.dataclass(frozen=True)
class ParsedELNURLFile:
    url: str
    title: typing.Optional[str]
    description: typing.Optional[str]
    user_id: typing.Optional[str]
    date_created: typing.Optional[datetime.datetime]


@dataclasses.dataclass(frozen=True)
class ParsedELNObject:
    id: str
    name: str
    url: typing.Optional[str]
    files: typing.List[typing.Union[ParsedELNURLFile, ParsedELNDatabaseFile]]
    versions: typing.List[ParsedELNVersion]
    comments: typing.List[ParsedELNComment]
    type: typing.Optional[str]
    type_id: typing.Optional[int]


@dataclasses.dataclass(frozen=True)
class ParsedELNUser:
    id: str
    name: typing.Optional[str]
    url: typing.Optional[str]


@dataclasses.dataclass(frozen=True)
class ParsedELNImport:
    objects: typing.List[ParsedELNObject]
    users: typing.List[ParsedELNUser]
    import_notes: typing.Dict[str, typing.List[str]]
    signed_by: typing.Optional[str]


def get_eln_import(eln_import_id: int) -> ELNImport:
    eln_import = eln_imports.ELNImport.query.filter_by(id=eln_import_id).first()
    if eln_import is None:
        raise errors.ELNImportDoesNotExistError()
    return ELNImport.from_database(eln_import)


def create_eln_import(
        user_id: int,
        file_name: str,
        zip_bytes: bytes
) -> ELNImport:
    eln_import = eln_imports.ELNImport(
        user_id=user_id,
        file_name=file_name,
        binary_data=zip_bytes,
        upload_utc_datetime=datetime.datetime.now(datetime.timezone.utc),
        import_utc_datetime=None,
    )
    db.session.add(eln_import)
    db.session.commit()
    return ELNImport.from_database(eln_import)


def delete_eln_import(
        eln_import_id: int
) -> None:
    eln_import = eln_imports.ELNImport.query.filter_by(id=eln_import_id).first()
    if eln_import is None:
        raise errors.ELNImportDoesNotExistError()
    db.session.delete(eln_import)
    db.session.commit()


def mark_eln_import_invalid(
        eln_import_id: int,
        invalid_reason: str
) -> None:
    eln_import = eln_imports.ELNImport.query.filter_by(id=eln_import_id).first()
    if eln_import is None:
        raise errors.ELNImportDoesNotExistError()
    eln_import.invalid_reason = invalid_reason
    db.session.add(eln_import)
    db.session.commit()


def _create_eln_import_action(
        action_type_id: int
) -> eln_imports.ELNImportAction:
    action = create_action(
        action_type_id=action_type_id,
        schema={
            'type': 'object',
            'title': {
                'en': 'Object Information'
            },
            'properties': {
                'name': {
                    'type': 'text',
                    'title': {
                        'en': 'Name'
                    }
                }
            },
            'required': ['name'],
            'propertyOrder': ['name']
        },
        is_hidden=True,
        disable_create_objects=True
    )
    set_action_translation(
        language_id=Language.ENGLISH,
        action_id=action.id,
        name='Import from .eln file',
        description='This action represents importing an object from an .eln file.',
        short_description='This action represents importing an object from an .eln file.'
    )
    set_action_translation(
        language_id=Language.GERMAN,
        action_id=action.id,
        name='Import aus .eln-Datei',
        description='Diese Aktion repräsentiert das Importieren eines Objekts aus einer .eln-Datei.',
        short_description='Diese Aktion repräsentiert das Importieren eines Objekts aus einer .eln-Datei.'
    )
    set_action_permissions_for_all_users(action_id=action.id, permissions=Permissions.READ)
    eln_import_action = eln_imports.ELNImportAction(action_id=action.id, action_type_id=action_type_id)
    db.session.add(eln_import_action)
    db.session.commit()
    return eln_import_action


def import_eln_file(
        eln_import_id: int,
        action_type_ids: typing.Optional[typing.List[typing.Optional[int]]] = None
) -> typing.Tuple[typing.List[int], typing.Dict[typing.Optional[str], User], typing.List[str]]:
    errors = []

    eln_import = eln_imports.ELNImport.query.filter_by(id=eln_import_id).first()
    if eln_import is None:
        return [], {}, ['Unknown ELN import']
    if eln_import.import_utc_datetime is not None:
        return [], {}, ['This ELN file has already been imported']
    user_id = eln_import.user_id

    parsed_data = parse_eln_file(eln_import_id=eln_import_id)
    if action_type_ids is None:
        action_type_ids = []
        for object_info in parsed_data.objects:
            action_type_ids.append(object_info.type_id)
    elif len(action_type_ids) != len(parsed_data.objects):
        return [], {}, ['Invalid Action Type ID information for this ELN file']

    eln_import.import_utc_datetime = datetime.datetime.now(datetime.timezone.utc)
    eln_import.signed_by = parsed_data.signed_by
    db.session.add(eln_import)
    db.session.commit()

    users_by_id: typing.Dict[typing.Optional[str], User] = {}
    for user_info in parsed_data.users:
        user = create_user(
            name=user_info.name,
            email=None,
            type=UserType.ELN_IMPORT_USER,
            eln_import_id=eln_import_id,
            eln_object_id=user_info.id
        )
        set_user_hidden(user.id, True)
        users_by_id[user_info.id] = user
    has_unknown_user = False
    for object_info in parsed_data.objects:
        for file_info in object_info.files:
            if file_info.user_id is None:
                has_unknown_user = True
                break
        for comment_info in object_info.comments:
            if comment_info.author_id is None:
                has_unknown_user = True
                break
    if has_unknown_user:
        user = create_user(
            name='Unknown User',
            email=None,
            type=UserType.ELN_IMPORT_USER,
            eln_import_id=eln_import_id,
            eln_object_id=''
        )
        set_user_hidden(user.id, True)

    action_ids_by_action_type_id: typing.Dict[typing.Optional[int], typing.Optional[int]] = {
        None: None
    }
    for action_type_id in set(action_type_ids):
        if action_type_id is None:
            continue
        eln_import_action = eln_imports.ELNImportAction.query.filter_by(action_type_id=action_type_id).first()
        if eln_import_action is None:
            eln_import_action = _create_eln_import_action(action_type_id)
        action_ids_by_action_type_id[action_type_id] = eln_import_action.action_id
    imported_object_ids = []
    for object_index, object_data in enumerate(zip(parsed_data.objects, action_type_ids)):
        object_info, action_type_id = object_data
        versions_by_id = {
            version_info.version_id: version_info
            for version_info in object_info.versions
        }
        version_ids = list(sorted(versions_by_id.keys()))
        initial_version_info = versions_by_id[version_ids[0]]
        try:
            imported_object = create_eln_import_object(
                importing_user_id=user_id,
                user_id=users_by_id[initial_version_info.user_id].id if initial_version_info.user_id else None,
                action_id=action_ids_by_action_type_id[action_type_id],
                data=initial_version_info.data,
                schema=initial_version_info.schema,
                eln_import_id=eln_import_id,
                eln_object_id=object_info.id,
                utc_datetime=initial_version_info.date_created,
                data_validator_arguments={'file_names_by_id': None, 'strict': False, 'allow_disabled_languages': True}
            )
        except Exception:
            errors.append(f'Failed to import object #{object_index}')
            continue
        if initial_version_info.user_id and initial_version_info.date_created:
            object_log.create_object(
                user_id=users_by_id[initial_version_info.user_id].id,
                object_id=imported_object.object_id,
                utc_datetime=initial_version_info.date_created,
                is_imported=True
            )
        db.session.add(eln_imports.ELNImportObject(
            object_id=imported_object.id,
            eln_import_id=eln_import_id,
            url=object_info.url
        ))
        db.session.commit()
        for version_index, version_id in enumerate(version_ids[1:], start=1):
            version_info = versions_by_id[version_id]
            try:
                update_object(
                    object_id=imported_object.id,
                    data=version_info.data,
                    schema=version_info.schema,
                    user_id=users_by_id[version_info.user_id].id if version_info.user_id else user_id,
                    data_validator_arguments={'file_names_by_id': None, 'strict': False, 'allow_disabled_languages': True},
                    create_log_entries=False,
                    utc_datetime=version_info.date_created
                )
            except Exception:
                errors.append(f'Failed to import object #{object_index} version #{version_index}')
            else:
                if version_info.user_id and version_info.date_created:
                    object_log.edit_object(
                        user_id=users_by_id[version_info.user_id].id,
                        object_id=imported_object.object_id,
                        version_id=version_index,
                        utc_datetime=version_info.date_created,
                        is_imported=True
                    )
        imported_object_ids.append(imported_object.id)
        for file_index, file_info in enumerate(object_info.files):
            file_user_id = user_id
            if isinstance(file_info, ParsedELNDatabaseFile):
                content = file_info.content
                if file_info.user_id:
                    file_user_id = users_by_id[file_info.user_id].id

                def save_content(file: typing.BinaryIO, content: bytes = content) -> None:
                    file.write(content)
                file_id = create_database_file(
                    object_id=imported_object.id,
                    user_id=file_user_id,
                    file_name=file_info.name,
                    save_content=save_content,
                    create_log_entry=False,
                    utc_datetime=file_info.date_created
                ).id
            elif isinstance(file_info, ParsedELNURLFile):
                if file_info.user_id:
                    file_user_id = users_by_id[file_info.user_id].id

                file_id = create_url_file(
                    object_id=imported_object.id,
                    user_id=file_user_id,
                    url=file_info.url,
                    create_log_entry=False,
                    utc_datetime=file_info.date_created
                ).id
            else:
                errors.append(f'Failed to import file #{file_index} for object #{object_index}')
                continue
            if file_info.user_id and file_info.date_created:
                object_log.upload_file(
                    user_id=file_user_id,
                    object_id=imported_object.object_id,
                    file_id=file_id,
                    utc_datetime=file_info.date_created,
                    is_imported=True
                )
            update_file_information(
                object_id=imported_object.id,
                file_id=file_id,
                user_id=file_user_id,
                title=file_info.title or '',
                description=file_info.description or ''
            )
        for comment_info in object_info.comments:
            comment_id = create_comment(
                object_id=imported_object.id,
                user_id=users_by_id[comment_info.author_id].id,
                content=comment_info.text,
                utc_datetime=comment_info.date_created,
                create_log_entry=False
            )
            object_log.post_comment(
                user_id=users_by_id[comment_info.author_id].id,
                object_id=imported_object.object_id,
                comment_id=comment_id,
                utc_datetime=comment_info.date_created,
                is_imported=True
            )
    return imported_object_ids, users_by_id, errors


def _eln_assert(assertion: bool, message: str = 'Invalid .eln file') -> None:
    if not assertion:
        raise errors.InvalidELNFileError(message)


def _json_contains_no_invalid_data(
        json_data: typing.Any
) -> bool:
    """
    Ensure that JSON data contains no invalid data such as strings with NULL bytes.
    """
    if isinstance(json_data, list):
        return all(_json_contains_no_invalid_data(item) for item in json_data)
    if isinstance(json_data, dict):
        return all(_json_contains_no_invalid_data(value) for value in json_data.values())
    if isinstance(json_data, str):
        if '\0' in json_data:
            return False
        return True
    return True


def _parse_eln_import_datetime(datetime_string: str) -> typing.Optional[datetime.datetime]:
    for datetime_format in [
        '%Y-%m-%dT%H:%M:%S.%f%z',
        '%Y-%m-%dT%H:%M:%S%z',
        '%Y-%m-%dT%H:%M:%S.%f',
        '%Y-%m-%dT%H:%M:%S',
        '%Y-%m-%d %H:%M:%S.%f',
        '%Y-%m-%d %H:%M:%S',
        '%Y-%m-%d',
    ]:
        try:
            parsed_datetime = datetime.datetime.strptime(datetime_string, datetime_format)
            if parsed_datetime.tzinfo is None:
                utc_datetime = parsed_datetime.replace(tzinfo=datetime.timezone.utc)
            else:
                utc_datetime = parsed_datetime.astimezone(tz=datetime.timezone.utc)
            return utc_datetime
        except Exception:
            continue
    return None


def _parse_author_ref(
        author_ref: typing.Any,
        graph_nodes_by_id: typing.Dict[str, typing.Any]
) -> typing.Optional[str]:
    return _parse_person_ref(
        person_ref=author_ref,
        graph_nodes_by_id=graph_nodes_by_id,
        node_name='author'
    )


def _parse_creator_ref(
        creator_ref: typing.Any,
        graph_nodes_by_id: typing.Dict[str, typing.Any]
) -> typing.Optional[str]:
    return _parse_person_ref(
        person_ref=creator_ref,
        graph_nodes_by_id=graph_nodes_by_id,
        node_name='creator'
    )


def _parse_person_ref(
        person_ref: typing.Any,
        graph_nodes_by_id: typing.Dict[str, typing.Any],
        node_name: str
) -> typing.Optional[str]:
    if person_ref is None:
        return None
    person_id: str
    _eln_assert(isinstance(person_ref, dict), f"Invalid {node_name} reference or node")
    person_ref = typing.cast(typing.Dict[str, typing.Any], person_ref)
    if '@id' in person_ref and person_ref['@id'] in graph_nodes_by_id:
        person_id = person_ref['@id']
        if graph_nodes_by_id[person_id] != person_ref:
            _eln_assert(list(person_ref.keys()) == ['@id'], f"Invalid {node_name} reference")
        person_node = graph_nodes_by_id[person_id]
    else:
        person_node = person_ref
        if '@id' in person_node:
            person_id = person_node['@id']
        else:
            _eln_assert('identifier' in person_node and isinstance(person_node['identifier'], str), f"Invalid {node_name} node")
            person_id = person_node['identifier']
            person_node['@id'] = person_id
        if person_id in graph_nodes_by_id:
            pass
            # TODO: the following check should be performed but is currently omitted to allow importing elabFTW exports, which may contain conflicting user information
            # _eln_assert(graph_nodes_by_id[person_id] == person_node, f"Invalid {node_name} reference")
        else:
            graph_nodes_by_id[person_id] = person_node
    _eln_assert(person_node['@type'] == 'Person', "Reference to node of wrong type")
    return person_id


def _json_has_valid_signature(json_bytes: bytes, signature: minisign.Signature) -> typing.Tuple[bool, typing.Optional[str]]:
    valid_schemes = ["https"]
    if flask.current_app.config["ELN_FILE_IMPORT_ALLOW_HTTP"]:
        valid_schemes.append("http")

    url = signature.trusted_comment

    try:
        split_url = urlsplit(url)
    except ValueError:
        return False, gettext("The signature does not contain a valid URL.")

    if split_url.scheme == "":
        return False, gettext("Missing URL scheme")
    if split_url.scheme not in valid_schemes:
        return False, gettext("Invalid URL scheme: %(scheme)s", scheme=split_url.scheme)
    if split_url.path != "/.well-known/keys.json":
        return False, gettext("The signature does not contain a valid URL pointing to /.well-known/keys.json.")

    try:
        res = requests.get(url, timeout=3)
    except requests.exceptions.ConnectionError:
        return False, gettext("Failed to connect to %(url)s.", url=url)

    if res.status_code != 200:
        return False, gettext("Received an error from %(url)s.", url=url)

    try:
        key_list = res.json()
    except requests.exceptions.JSONDecodeError:
        return False, gettext("%(url)s did not return a valid JSON object.", url=url)

    if isinstance(key_list, dict):
        key_list = [key_list]
    if not isinstance(key_list, list):
        return False, gettext("%(url)s did not return a valid JSON list.", url=url)

    num_failed_keys = 0
    for elem in key_list:
        if not isinstance(elem, dict):
            num_failed_keys += 1
            continue
        if "contentUrl" in elem:
            try:
                response = requests.get(elem['contentUrl'], timeout=3)
            except requests.exceptions.ConnectionError:
                num_failed_keys += 1
                continue
        else:
            num_failed_keys += 1
            continue

        if response.status_code != 200:
            num_failed_keys += 1
            continue

        try:
            pub = minisign.PublicKey.from_base64(response.content)
        except binascii.Error:
            # not in base64 format
            num_failed_keys += 1
            continue
        except ValueError:
            # invalid content, e.g. wrong bytes specifying minisign SignatureAlgorithm
            num_failed_keys += 1
            continue

        try:
            pub.verify(json_bytes, signature)
            return True, None
        except minisign.exceptions.VerifyError:
            continue

    if num_failed_keys > 0:
        return False, gettext(
            "There was an error checking %(num_failed_keys)s of %(num_keys)s public keys. No matching public key was found.",
            num_failed_keys=num_failed_keys,
            num_keys=len(key_list)
        )
    return False, gettext(
        "Successfully fetched all %(num_keys)s keys but no matching public key was found.",
        num_keys=len(key_list)
    )


def parse_eln_file(
        eln_import_id: int
) -> ParsedELNImport:

    eln_import = eln_imports.ELNImport.query.filter_by(id=eln_import_id).first()
    if eln_import is None:
        raise errors.ELNImportDoesNotExistError()
    zip_bytes = eln_import.binary_data

    parsed_data = ParsedELNImport(
        objects=[],
        users=[],
        import_notes={},
        signed_by=None
    )

    try:
        with zipfile.ZipFile(io.BytesIO(zip_bytes)) as zip_file:
            _eln_assert(zip_file.testzip() is None, ".eln file must be valid .zip file")

            member_names = {
                os.path.normpath(member_name): member_name
                for member_name in zip_file.namelist()
            }
            root_path_names = set()
            for member_name in member_names:
                path_name, file_name = os.path.split(member_name)
                root_path_names.add(path_name.split(os.sep)[0])
            _eln_assert(len(root_path_names) == 1, ".eln file must contain a single root directory")
            root_path_name = list(root_path_names)[0]

            ro_crate_metadata_file_name = root_path_name + '/ro-crate-metadata.json'
            ro_crate_metadata_file_name = os.path.normpath(ro_crate_metadata_file_name)
            _eln_assert(ro_crate_metadata_file_name in member_names, ".eln file must contain ro-crate-metadata.json in its root directory")
            try:
                with zip_file.open(member_names[ro_crate_metadata_file_name]) as ro_crate_metadata_file:
                    ro_crate_metadata = json.load(ro_crate_metadata_file)
            except Exception:
                raise errors.InvalidELNFileError("ro-crate-metadata.json must contain a JSON-encoded object")

            with zip_file.open(root_path_name + '/ro-crate-metadata.json') as ro_crate_metadata_file:
                ro_crate_metadata_bytes = ro_crate_metadata_file.read()

            ro_crate_metadata_sig_file_name = root_path_name + '/ro-crate-metadata.json.minisig'
            ro_crate_metadata_sig_file_name = os.path.normpath(ro_crate_metadata_sig_file_name)
            if ro_crate_metadata_sig_file_name in member_names:
                skip_validation = False
                try:
                    with zip_file.open(member_names[ro_crate_metadata_sig_file_name]) as ro_crate_metadata_sig_bytes:
                        ro_crate_metadata_sig = minisign.Signature.from_bytes(ro_crate_metadata_sig_bytes.read())
                except minisign.ParseError:
                    raise errors.InvalidELNFileError("ro-crate-metadata.json.minisig must contain a minisign signature")
                except Exception:
                    skip_validation = True

                if not skip_validation:
                    has_valid_signature, note = _json_has_valid_signature(ro_crate_metadata_bytes, ro_crate_metadata_sig)
                    if not has_valid_signature and note is not None:
                        parsed_data.import_notes["Signature"] = [note]
                    else:
                        parsed_data = ParsedELNImport(
                            parsed_data.objects,
                            parsed_data.users,
                            parsed_data.import_notes,
                            ro_crate_metadata_sig.trusted_comment.removesuffix("/.well-known/keys.json") if has_valid_signature else None
                        )
            _eln_assert(isinstance(ro_crate_metadata, dict), "ro-crate-metadata.json must contain a JSON-encoded object")
            _eln_assert(_json_contains_no_invalid_data(ro_crate_metadata), "ro-crate-metadata.json must not contain invalid data")
            supported_ro_crate_versions = [
                '1.0',
                '1.1',
                '1.2',
            ]
            _eln_assert(
                any(
                    ro_crate_metadata.get('@context') == f'https://w3id.org/ro/crate/{supported_ro_crate_version}/context'
                    for supported_ro_crate_version in supported_ro_crate_versions
                ),
                f"ro-crate-metadata.json @context must be RO-Crate context for RO-Crate version {', '.join(supported_ro_crate_versions[:1])} or {supported_ro_crate_versions[-1]}"
            )

            _eln_assert(isinstance(ro_crate_metadata.get('@graph'), list), "ro-crate-metadata.json @graph must be list")
            graph_nodes_by_id: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
            for graph_node in ro_crate_metadata['@graph']:
                _eln_assert(isinstance(graph_node, dict), "ro-crate-metadata.json @graph entries must be valid objects")
                _eln_assert(all(isinstance(key, str) for key in graph_node), "ro-crate-metadata.json @graph entries must be valid objects")
                _eln_assert(isinstance(graph_node.get('@id'), str), "ro-crate-metadata.json @graph entries must have string @id")
                _eln_assert(isinstance(graph_node.get('@type'), str) or (isinstance(graph_node.get('@type'), list) and len(graph_node['@type']) >= 1 and all(isinstance(type_entry, str) for type_entry in graph_node['@type'])), "ro-crate-metadata.json @graph entries must have string or list of strings @type")
                _eln_assert(graph_node['@id'] not in graph_nodes_by_id, "ro-crate-metadata.json @graph entries must not have duplicate @id values")
                graph_nodes_by_id[graph_node['@id']] = graph_node

            _eln_assert('./' in graph_nodes_by_id, "ro-crate-metadata.json @graph must contain entry with @id value './'")
            root_node = graph_nodes_by_id['./']
            _eln_assert('ro-crate-metadata.json' in graph_nodes_by_id, "ro-crate-metadata.json @graph must contain entry with @id value 'ro-crate-metadata.json'")
            metadata_node = graph_nodes_by_id['ro-crate-metadata.json']
            _eln_assert(root_node['@type'] == 'Dataset' or root_node['@type'] == ['Dataset'], "ro-crate-metadata.json @graph root node must be Dataset")
            if 'version' in metadata_node:
                _eln_assert(metadata_node['version'] in supported_ro_crate_versions or any(metadata_node['version'].startswith(supported_ro_crate_version + '.') for supported_ro_crate_version in supported_ro_crate_versions), f"ro-crate-metadata.json node has unsupported RO-Crate version (supported are: {', '.join(supported_ro_crate_versions)})")
            root_sdpublisher = root_node.get('sdPublisher')
            metadata_sdpublisher = metadata_node.get('sdPublisher')
            _eln_assert(isinstance(root_sdpublisher, dict) or isinstance(metadata_sdpublisher, dict), "ro-crate-metadata.json @graph root node or ro-crate-metadata.json node must contain valid sdPublisher")
            _eln_assert(isinstance(root_sdpublisher, dict) or root_sdpublisher is None, "ro-crate-metadata.json @graph root node sdPublisher is invalid")
            _eln_assert(isinstance(metadata_sdpublisher, dict) or metadata_sdpublisher is None, "ro-crate-metadata.json @graph ro-crate-metadata.json node sdPublisher is invalid")
            if root_sdpublisher is not None and metadata_sdpublisher is not None:
                _eln_assert(root_sdpublisher == metadata_sdpublisher, "ro-crate-metadata.json @graph contains conflicting sdPublisher information")
            if root_sdpublisher is not None:
                sdpublisher = root_sdpublisher
            else:
                sdpublisher = metadata_sdpublisher
            _eln_assert(all(isinstance(key, str) for key in sdpublisher), "ro-crate-metadata.json @graph sdPublisher is invalid")
            if set(sdpublisher.keys()) == {'@id'}:
                _eln_assert(isinstance(sdpublisher['@id'], str), "ro-crate-metadata.json @graph sdPublisher is invalid")
                sdpublisher = graph_nodes_by_id.get(sdpublisher['@id'])
                _eln_assert(isinstance(sdpublisher, dict), "ro-crate-metadata.json @graph sdPublisher is invalid")
                sdpublisher = typing.cast(typing.Dict[str, typing.Any], sdpublisher)
                _eln_assert(all(isinstance(key, str) for key in sdpublisher), "ro-crate-metadata.json @graph sdPublisher is invalid")
            _eln_assert(sdpublisher.get('@type') == 'Organization', "ro-crate-metadata.json @graph sdPublisher is invalid")
            if sdpublisher.get('name') == 'SampleDB':
                eln_dialect = 'SampleDB'
            else:
                eln_dialect = None
            _eln_assert('hasPart' in root_node, "ro-crate-metadata.json @graph root node must have parts")
            if not isinstance(root_node['hasPart'], list):
                root_node['hasPart'] = [root_node['hasPart']]

            for object_node_ref in root_node['hasPart']:
                _eln_assert(isinstance(object_node_ref, dict), "Invalid reference")
                _eln_assert(list(object_node_ref.keys()) == ['@id'], "Invalid reference")
                _eln_assert(object_node_ref['@id'] in graph_nodes_by_id, "Reference to unknown ID")

                object_node = graph_nodes_by_id[object_node_ref['@id']]
                _eln_assert(isinstance(object_node.get('name'), str), "Missing name for Dataset")
                name = object_node['name']

                parsed_data.import_notes[object_node["@id"]] = []

                _eln_assert(isinstance(object_node.get('description', ''), (str, dict)), "Invalid description for Dataset")
                description = object_node.get('description', '')
                if isinstance(description, dict):
                    if list(description.keys()) == ['@id']:
                        _eln_assert(description['@id'] in graph_nodes_by_id, "Reference to unknown ID")
                        description = graph_nodes_by_id[description['@id']]
                    _eln_assert(isinstance(object_node_ref, dict), "Invalid description")
                    _eln_assert(description.get('@type') == 'TextObject', "Invalid description for Dataset")
                    description_is_markdown = description.get('encodingFormat') == 'text/markdown'
                    description = description.get('text', '')
                else:
                    description_is_markdown = False

                if 'url' in object_node:
                    _eln_assert(isinstance(object_node.get('url'), str), "Invalid URL for Dataset")
                    try:
                        url = object_node['url']
                        parse_url(url, valid_schemes=('http', 'https'))
                    except Exception:
                        raise errors.InvalidELNFileError("Invalid URL for Dataset")
                else:
                    url = None
                if eln_dialect == 'SampleDB':
                    _eln_assert(object_node_ref['@id'].startswith('./objects/'), "Invalid @id for Dataset for SampleDB .eln file")
                    try:
                        original_object_id_str = object_node_ref['@id'][len('./objects/'):].rstrip('/')
                        original_object_id = int(original_object_id_str)
                        _eln_assert(str(original_object_id) == original_object_id_str and original_object_id > 0, "Invalid @id for Dataset for SampleDB .eln file")
                    except ValueError:
                        _eln_assert(False, "Invalid @id for Dataset for SampleDB .eln file")
                    _eln_assert(isinstance(url, str), "Invalid url for Dataset for SampleDB .eln file")
                    _eln_assert(url.endswith(object_node_ref['@id'].rstrip('/')[1:]), "Invalid url for Dataset for SampleDB .eln file")
                    eln_source_url = url[:-len(object_node_ref['@id'][2:])]
                else:
                    eln_source_url = None

                if object_node.get('genre') in {'sample', 'measurement', 'simulation', 'experiment', 'procedure'}:
                    object_type = {
                        'experiment': 'measurement',
                        'procedure': 'measurement'
                    }.get(object_node['genre'], object_node['genre'])
                    object_type_id = {
                        'sample': ActionType.SAMPLE_CREATION,
                        'measurement': ActionType.MEASUREMENT,
                        'simulation': ActionType.SIMULATION,
                    }.get(object_type)
                else:
                    object_type = None
                    object_type_id = None

                if 'comment' in object_node and not isinstance(object_node['comment'], list):
                    object_node['comment'] = [object_node['comment']]
                comments: typing.List[ParsedELNComment] = []
                for comment_ref in object_node.get('comment', []):
                    _eln_assert(isinstance(comment_ref, dict), "Invalid comment reference or node")
                    if '@id' in comment_ref and comment_ref['@id'] in graph_nodes_by_id:
                        _eln_assert(list(comment_ref.keys()) == ['@id'], "Invalid comment reference")
                        comment_node = graph_nodes_by_id[comment_ref['@id']]
                    else:
                        comment_node = comment_ref
                    author_ref = comment_node.get('author')
                    if author_ref is None:
                        author_id = None
                    else:
                        _eln_assert(isinstance(author_ref, dict), "Invalid author reference or node")
                        author_ref = typing.cast(typing.Dict[str, typing.Any], author_ref)
                        if '@id' in author_ref and author_ref['@id'] in graph_nodes_by_id:
                            author_id = author_ref['@id']
                            if graph_nodes_by_id[author_id] != author_ref:
                                _eln_assert(list(author_ref.keys()) == ['@id'], "Invalid author reference")
                            author_node = graph_nodes_by_id[author_id]
                        else:
                            author_node = author_ref
                            if '@id' in author_node:
                                author_id = author_node['@id']
                            else:
                                _eln_assert('identifier' in author_node, "Invalid author node")
                                author_id = author_node['identifier']
                                author_node['@id'] = author_id
                            if author_id in graph_nodes_by_id:
                                pass
                                # TODO: the following check should be performed but is currently omitted to allow importing elabFTW exports, which may contain conflicting user information
                                # _eln_assert(graph_nodes_by_id[author_id] == author_node, "Invalid author reference")
                            else:
                                graph_nodes_by_id[author_id] = author_node
                        _eln_assert(author_node['@type'] == 'Person', "Reference to node of wrong type")
                    _eln_assert(isinstance(comment_node.get('text'), str), "Invalid text for Comment")
                    _eln_assert(isinstance(comment_node.get('dateCreated'), str), "Invalid dateCreated for Comment")
                    date_created: typing.Optional[datetime.datetime]
                    try:
                        date_created = datetime.datetime.fromisoformat(comment_node['dateCreated'])
                    except Exception:
                        raise errors.InvalidELNFileError("Invalid dateCreated for Comment")
                    comments.append(ParsedELNComment(
                        text=comment_node['text'],
                        date_created=date_created,
                        author_id=author_id
                    ))

                files: typing.List[typing.Union[ParsedELNURLFile, ParsedELNDatabaseFile]] = []
                versions: typing.List[ParsedELNVersion] = []

                fallback_data = {
                    'name': {
                        '_type': 'text',
                        'text': {'en': name}
                    }
                }
                if description:
                    fallback_data['description'] = {
                        '_type': 'text',
                        'text': {'en': description},
                    }
                    if description_is_markdown:
                        fallback_data['description']['is_markdown'] = True  # type: ignore[assignment]
                fallback_schema = {
                    'type': 'object',
                    'title': {
                        'en': 'Object Information',
                        'de': 'Objektinformationen',
                    },
                    'properties': {
                        'name': {
                            'type': 'text',
                            'title': {
                                'en': 'Name',
                                'de': 'Name'
                            }
                        }
                    },
                    'required': ['name'],
                    'propertyOrder': ['name']
                }
                fallback_schema_properties = typing.cast(typing.Dict[str, typing.Any], fallback_schema['properties'])
                fallback_schema_property_order = typing.cast(typing.List[str], fallback_schema['propertyOrder'])
                if description:
                    fallback_schema_properties['description'] = {
                        'type': 'text',
                        'title': {
                            'en': 'Description',
                            'de': 'Beschreibung'
                        },
                        'multiline': '\n' in description
                    }
                    if description_is_markdown:
                        fallback_schema_properties['description']['multiline'] = False
                        fallback_schema_properties['description']['markdown'] = True
                    fallback_schema_property_order.append('description')
                has_metadata = False
                if 'keywords' in object_node:
                    fallback_schema_properties['tags'] = {
                        'type': 'tags',
                        'title': {
                            'en': 'Tags',
                            'de': 'Tags'
                        }
                    }
                    fallback_schema_property_order.append('tags')
                    if isinstance(object_node['keywords'], list):
                        keywords = [
                            keyword
                            for keyword in object_node['keywords']
                            if isinstance(keyword, str)
                        ]
                    elif isinstance(object_node['keywords'], str):
                        keywords = [
                            keyword
                            for keyword in object_node['keywords'].split(',')
                        ]
                    else:
                        keywords = []
                    tags = []
                    for keyword in keywords:
                        keyword = keyword.strip().lower()
                        # skip duplicate tags
                        if keyword in tags:
                            continue
                        # skip tags with invalid characters
                        if any(c not in 'abcdefghijklmnopqrstuvwxyz0123456789_-äöüß' for c in keyword):
                            continue
                        # skip numeric tags unless enabled
                        if all(c in string.digits for c in keyword) and not flask.current_app.config['ENABLE_NUMERIC_TAGS']:
                            continue
                        tags.append(keyword)
                    fallback_data['tags'] = {
                        '_type': 'tags',
                        'tags': tags
                    }
                else:
                    tags = []
                if 'variableMeasured' in object_node and object_node['variableMeasured'] is not None:
                    if not isinstance(object_node['variableMeasured'], list):
                        object_node['variableMeasured'] = [object_node['variableMeasured']]
                    has_metadata = True
                    property_values = []
                    for property_value in object_node['variableMeasured']:
                        _eln_assert(isinstance(property_value, dict), "Invalid variableMeasured item for Dataset")
                        if '@id' in property_value and property_value['@id'] in graph_nodes_by_id:
                            property_value = graph_nodes_by_id[property_value['@id']]
                        # skip untyped variableMeasured entries, e.g. used by elabFTW for elabftw_metadata
                        if '@type' not in property_value:
                            continue
                        _eln_assert(property_value.get('@type') == 'PropertyValue', "Invalid variableMeasured item type for Dataset")
                        _eln_assert(isinstance(property_value.get('propertyID'), str), "Invalid variableMeasured item propertyID for Dataset")
                        _eln_assert(isinstance(property_value.get('name', ''), str), "Invalid variableMeasured item name for Dataset")
                        _eln_assert(isinstance(property_value.get('value'), (str, bool, int, float)), "Invalid variableMeasured item value for Dataset")
                        _eln_assert(isinstance(property_value.get('unitCode', ''), str), "Invalid variableMeasured item unitCode for Dataset")
                        _eln_assert(isinstance(property_value.get('unitText', ''), str), "Invalid variableMeasured item unitText for Dataset")
                        property_values.append(property_value)
                    fallback_data, fallback_schema = _convert_property_values_to_data_and_schema(
                        property_values=property_values,
                        name=name,
                        description=description,
                        description_is_markdown=description_is_markdown,
                        tags=tags
                    )
                if 'hasPart' in object_node and not isinstance(object_node['hasPart'], list):
                    object_node['hasPart'] = [object_node['hasPart']]
                for object_part_ref in object_node.get('hasPart', []):
                    _eln_assert(isinstance(object_part_ref, dict), "Invalid reference")
                    _eln_assert(list(object_part_ref.keys()) == ['@id'], "Invalid reference")
                    _eln_assert(object_part_ref['@id'] in graph_nodes_by_id, "Reference to unknown ID")
                    object_part = graph_nodes_by_id[object_part_ref['@id']]
                    if object_part['@type'] == 'File':
                        _eln_assert(isinstance(object_part.get('name'), str), "Invalid name for File")
                        if 'description' in object_part:
                            description = object_part.get('description')
                            if isinstance(description, dict):
                                if list(description.keys()) == ['@id']:
                                    _eln_assert(description['@id'] in graph_nodes_by_id, "Reference to unknown ID")
                                    description = graph_nodes_by_id[description['@id']]
                                _eln_assert(isinstance(object_node_ref, dict), "Invalid description")
                                _eln_assert(description.get('@type') == 'TextObject', "Invalid description for File")
                                description = description.get('text', '')
                            else:
                                _eln_assert(isinstance(object_part.get('description'), str), "Invalid description for File")
                        file_path = object_part['@id']
                        try:
                            file_url_parse_result = urllib.parse.urlparse(file_path)
                        except ValueError:
                            file_url_parse_result = None
                        if file_url_parse_result is not None and file_url_parse_result.netloc and file_url_parse_result.scheme in ('http', 'https'):
                            file_url = object_part.get('url', file_path)
                            try:
                                parse_url(file_url, valid_schemes=('http', 'https'))
                            except Exception:
                                parsed_data.import_notes[object_node["@id"]].append(gettext('Failed to import %(file_url)s as URL file', file_url=file_url))
                            else:
                                date_created = None
                                if 'dateCreated' in object_part:
                                    _eln_assert(isinstance(object_part.get('dateCreated'), str), "Invalid dateCreated for File")
                                    date_created = _parse_eln_import_datetime(object_part['dateCreated'])
                                author_ref = object_part.get('author')
                                author_id = _parse_author_ref(author_ref, graph_nodes_by_id)
                                files.append(ParsedELNURLFile(
                                    url=file_url,
                                    title=object_part.get('name'),
                                    description=description,
                                    user_id=author_id,
                                    date_created=date_created
                                ))
                            continue
                        if file_path.startswith('./'):
                            file_path = file_path[1:]
                        elif not file_path.startswith('/'):
                            file_path = '/' + file_path
                        file_path = root_path_name + file_path
                        file_path = os.path.normpath(file_path)
                        _eln_assert(file_path in member_names, "File not found in .eln file (1)")
                        file_name = os.path.split(file_path)[1]
                        with zip_file.open(member_names[file_path]) as file:
                            file_data = file.read()
                        if isinstance(object_part.get('contentSize'), str):
                            try:
                                object_part['contentSize'] = int(object_part['contentSize'])
                            except ValueError:
                                pass
                        _eln_assert(isinstance(object_part.get('contentSize'), int), "Invalid content size for File")
                        _eln_assert(len(file_data) == object_part['contentSize'], "Content size mismatch for File")
                        if 'sha256' in object_part:
                            _eln_assert(isinstance(object_part.get('sha256'), str), "Invalid SHA256 hash for File")
                            _eln_assert(hashlib.sha256(file_data).hexdigest() == object_part['sha256'], "Hash mismatch for File")
                        if eln_dialect == 'SampleDB' and object_part['@id'] == object_node_ref['@id'] + ('/' if not object_node_ref['@id'].endswith('/') else '') + 'files.json':
                            try:
                                files_info = json.loads(file_data.decode('utf-8'))
                            except Exception:
                                raise errors.InvalidELNFileError("Invalid files.json for SampleDB .eln file")
                            _eln_assert(isinstance(files_info, list), "Invalid files.json for SampleDB .eln file")
                            for file_info in files_info:
                                _eln_assert(isinstance(file_info, dict), "Invalid files.json for SampleDB .eln file")
                                if 'url' not in file_info:
                                    continue
                                _eln_assert(isinstance(file_info.get('url'), str), "Invalid URL file for SampleDB .eln file")
                                _eln_assert(isinstance(file_info.get('title'), str) or file_info.get('title') is None, "Invalid URL file for SampleDB .eln file")
                                _eln_assert(isinstance(file_info.get('description'), str) or file_info.get('description') is None, "Invalid URL file for SampleDB .eln file")
                                _eln_assert(isinstance(file_info.get('uploader_id'), int), "Invalid URL file for SampleDB .eln file")
                                author_sampledb_id = file_info.get('uploader_id')
                                try:
                                    file_url = file_info['url']
                                    parse_url(file_url, valid_schemes=('http', 'https'))
                                except Exception:
                                    raise errors.InvalidELNFileError("Invalid URL file for SampleDB .eln file")
                                try:
                                    date_created = datetime.datetime.strptime(file_info.get('utc_datetime', ''), '%Y-%m-%dT%H:%M:%S.%f')
                                    date_created = date_created.replace(tzinfo=datetime.timezone.utc)
                                except Exception:
                                    date_created = None
                                files.append(ParsedELNURLFile(
                                    url=file_url,
                                    title=file_info.get('title'),
                                    description=file_info.get('description'),
                                    user_id=f'./users/{author_sampledb_id}',
                                    date_created=date_created
                                ))
                        else:
                            date_created = None
                            if 'dateCreated' in object_part:
                                _eln_assert(isinstance(object_part.get('dateCreated'), str), "Invalid dateCreated for File")
                                date_created = _parse_eln_import_datetime(object_part['dateCreated'])
                            author_ref = object_part.get('author')
                            author_id = _parse_author_ref(author_ref, graph_nodes_by_id)
                            files.append(ParsedELNDatabaseFile(
                                name=file_name,
                                title=object_part['name'],
                                description=object_part.get('description'),
                                content=file_data,
                                user_id=author_id,
                                date_created=date_created,
                            ))
                    if eln_dialect == 'SampleDB' and object_part['@type'] == 'Dataset':
                        _eln_assert(any(object_part['@id'].startswith(object_node['@id'] + ('/' if not object_node['@id'].endswith('/') else '') + version_suffix) for version_suffix in ('version/', 'versions/')), "SampleDB .eln file must only contain versions as Dataset parts of objects")
                        try:
                            version_id = int(object_part['@id'].strip('/').rsplit('/', maxsplit=1)[1])
                        except ValueError:
                            raise errors.InvalidELNFileError("SampleDB .eln file must only contain versions as Dataset parts of objects")
                        _eln_assert(isinstance(object_part.get('hasPart'), list), "SampleDB .eln file must contain data and schema for each version")
                        _eln_assert(len(object_part['hasPart']) == 2, "SampleDB .eln file must contain data and schema for each version")
                        _eln_assert({'@id': object_part['@id'] + ('/' if not object_part['@id'].endswith('/') else '') + 'data.json'} in object_part['hasPart'], "SampleDB .eln file must contain data and schema for each version")
                        _eln_assert({'@id': object_part['@id'] + ('/' if not object_part['@id'].endswith('/') else '') + 'schema.json'} in object_part['hasPart'], "SampleDB .eln file must contain data and schema for each version")
                        data_file_name = object_part['@id'] + '/data.json'
                        if data_file_name.startswith('./'):
                            data_file_name = data_file_name[1:]
                        elif not data_file_name.startswith('/'):
                            data_file_name = '/' + data_file_name
                        data_file_name = root_path_name + data_file_name
                        data_file_name = os.path.normpath(data_file_name)
                        _eln_assert(data_file_name in member_names, "SampleDB .eln file must contain data and schema for each version")
                        schema_file_name = object_part['@id'] + '/schema.json'
                        if schema_file_name.startswith('./'):
                            schema_file_name = schema_file_name[1:]
                        elif not schema_file_name.startswith('/'):
                            schema_file_name = '/' + schema_file_name
                        schema_file_name = root_path_name + schema_file_name
                        schema_file_name = os.path.normpath(schema_file_name)
                        _eln_assert(schema_file_name in member_names, "SampleDB .eln file must contain data and schema for each version")
                        try:
                            with zip_file.open(member_names[data_file_name]) as data_file:
                                data_file_content = data_file.read()
                                _eln_assert(isinstance(graph_nodes_by_id[object_part['@id'] + 'data.json'].get('sha256'), str), "Invalid SHA256 hash for data.json")
                                _eln_assert(hashlib.sha256(data_file_content).hexdigest() == graph_nodes_by_id[object_part['@id'] + 'data.json']['sha256'], "Hash mismatch for data.json")
                                version_data = json.loads(data_file_content.decode('utf-8'))
                            with zip_file.open(member_names[schema_file_name]) as schema_file:
                                schema_file_content = schema_file.read()
                                _eln_assert(isinstance(graph_nodes_by_id[object_part['@id'] + 'schema.json'].get('sha256'), str), "Invalid SHA256 hash for schema.json")
                                _eln_assert(hashlib.sha256(schema_file_content).hexdigest() == graph_nodes_by_id[object_part['@id'] + 'schema.json']['sha256'], "Hash mismatch for schema.json")
                                version_schema = json.loads(schema_file_content.decode('utf-8'))
                        except errors.InvalidELNFileError:
                            raise
                        except Exception:
                            raise errors.InvalidELNFileError("SampleDB .eln file must contain data and schema for each version")
                        creator_ref = object_part.get('creator')
                        if creator_ref is None and isinstance(object_part.get('author'), dict):
                            creator_ref = object_part.get('author')
                        creator_id = _parse_creator_ref(creator_ref, graph_nodes_by_id)
                        date_created = None
                        if 'dateCreated' in object_part:
                            _eln_assert(isinstance(object_part.get('dateCreated'), str), "Invalid dateCreated for Dataset")
                            date_created = _parse_eln_import_datetime(object_part['dateCreated'])
                        if version_schema is None or version_data is None:
                            versions.append(ParsedELNVersion(
                                version_id=version_id,
                                data=None,
                                schema=None,
                                user_id=creator_id,
                                date_created=date_created
                            ))
                            parsed_data.import_notes[object_node["@id"]].append(gettext('Missing data or schema for version %(version_id)s of object %(eln_object_id)s', version_id=version_id, eln_object_id=object_node['@id']))
                        else:
                            _remove_template_ids(version_schema)
                            _replace_references(version_data, eln_source_url)
                            try:
                                validate_schema(version_schema, strict=False)
                                validate(version_data, version_schema, allow_disabled_languages=True, strict=False, file_names_by_id=None)
                                versions.append(ParsedELNVersion(
                                    version_id=version_id,
                                    data=version_data,
                                    schema=version_schema,
                                    user_id=creator_id,
                                    date_created=date_created
                                ))
                            except errors.ValidationError as e:
                                versions.append(ParsedELNVersion(
                                    version_id=version_id,
                                    data=fallback_data,
                                    schema=fallback_schema,
                                    user_id=creator_id,
                                    date_created=date_created
                                ))
                                parsed_data.import_notes[object_node["@id"]].append(gettext('Invalid data or schema in version %(version_id)s of object %(eln_object_id)s (%(error)s)', version_id=version_id, eln_object_id=object_node['@id'], error=e))
                if not versions:
                    creator_ref = object_node.get('creator')
                    if creator_ref is None and isinstance(object_node.get('author'), dict):
                        creator_ref = object_node.get('author')
                    creator_id = _parse_creator_ref(creator_ref, graph_nodes_by_id)
                    date_created = None
                    if 'dateCreated' in object_node:
                        _eln_assert(isinstance(object_node.get('dateCreated'), str), "Invalid dateCreated for Dataset")
                        date_created = _parse_eln_import_datetime(object_node['dateCreated'])
                    if not has_metadata:
                        fallback_data['import_note'] = {
                            '_type': 'text',
                            'text': {
                                'en': 'The .eln file did not contain any valid flexible metadata for this object.',
                                'de': 'Die .eln-Datei enthielt keine gültigen flexiblen Metadaten für dieses Objekt.'
                            }
                        }
                        parsed_data.import_notes[object_node["@id"]].append(gettext('The .eln file did not contain any valid flexible metadata for this object.'))
                        fallback_schema_property_order.append('import_note')
                        fallback_schema_properties['import_note'] = {
                            'type': 'text',
                            'languages': ['en', 'de'],
                            'title': {
                                'en': 'Import Note',
                                'de': 'Import-Hinweis'
                            }
                        }
                    versions.append(ParsedELNVersion(
                        version_id=0,
                        data=fallback_data,
                        schema=fallback_schema,
                        user_id=creator_id,
                        date_created=date_created
                    ))
                parsed_data.objects.append(ParsedELNObject(
                    id=object_node_ref['@id'],
                    name=name,
                    url=url,
                    files=files,
                    versions=versions,
                    comments=comments,
                    type=object_type,
                    type_id=object_type_id,
                ))
            for graph_node in graph_nodes_by_id.values():
                if graph_node['@type'] == 'Person':
                    if 'name' in graph_node and graph_node['name'] is None:
                        name = None
                    else:
                        if 'name' in graph_node:
                            _eln_assert(isinstance(graph_node['name'], str), "Invalid name for Person")
                            name = graph_node['name'].strip()
                        elif 'familyName' in graph_node and 'givenName' in graph_node:
                            _eln_assert(isinstance(graph_node['familyName'], str), "Invalid family name for Person")
                            _eln_assert(isinstance(graph_node['givenName'], str), "Invalid given name for Person")
                            name = graph_node['givenName'] + ' ' + graph_node['familyName']
                            name = name.strip()
                        else:
                            name = None
                        _eln_assert(name, "Invalid name for Person")
                    url = None
                    if 'url' in graph_node:
                        _eln_assert(isinstance(graph_node['url'], str), "Invalid url for Person")
                        url = graph_node['url']
                    elif 'identifier' in graph_node:
                        _eln_assert(isinstance(graph_node['identifier'], str), "Invalid identifier for Person")
                        if graph_node['identifier'].startswith('https://orcid.org/'):
                            url = graph_node['identifier']
                    parsed_data.users.append(ParsedELNUser(
                        id=graph_node['@id'],
                        name=name,
                        url=url,
                    ))
    except errors.InvalidELNFileError:
        raise
    except zipfile.BadZipfile:
        raise errors.InvalidELNFileError(".eln file must be valid .zip file")
    except Exception as e:
        raise errors.InvalidELNFileError("Unknown error") from e

    if not parsed_data.objects:
        raise errors.InvalidELNFileError(".eln file must contain at least one object")
    return parsed_data


def remove_expired_eln_imports() -> None:
    expired_eln_imports = eln_imports.ELNImport.query.filter(db.and_(
        eln_imports.ELNImport.import_utc_datetime.is_(None),
        eln_imports.ELNImport.upload_utc_datetime < datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(days=1)
    )).all()
    if expired_eln_imports:
        for eln_import in expired_eln_imports:
            db.session.delete(eln_import)
        db.session.commit()


def get_pending_eln_imports(user_id: int) -> typing.List[ELNImport]:
    pending_eln_imports = eln_imports.ELNImport.query.filter_by(
        import_utc_datetime=None,
        user_id=user_id
    ).order_by(eln_imports.ELNImport.upload_utc_datetime).all()
    return [
        ELNImport.from_database(eln_import)
        for eln_import in pending_eln_imports
    ]


def get_eln_import_object_url(
        object_id: int
) -> typing.Optional[str]:
    eln_import_object = eln_imports.ELNImportObject.query.filter_by(object_id=object_id).first()
    if eln_import_object is None:
        return None
    return eln_import_object.url


def get_eln_import_for_object(
        object_id: int
) -> typing.Optional[typing.Tuple[ELNImport, typing.Optional[str]]]:
    eln_import_object = eln_imports.ELNImportObject.query.filter_by(object_id=object_id).first()
    if eln_import_object is None:
        return None
    return ELNImport.from_database(eln_import_object.eln_import), eln_import_object.url


def get_eln_import_objects(
        eln_import_id: int
) -> typing.Sequence[Object]:
    eln_import_objects = eln_imports.ELNImportObject.query.filter_by(eln_import_id=eln_import_id).all()
    return [
        get_object(eln_import_object.object_id)
        for eln_import_object in eln_import_objects
    ]


def get_eln_import_users(
        eln_import_id: int
) -> typing.Sequence[User]:
    eln_import_users = users.User.query.filter_by(type=UserType.ELN_IMPORT_USER, eln_import_id=eln_import_id).all()
    return [
        User.from_database(eln_import_user)
        for eln_import_user in eln_import_users
    ]


def get_import_signed_by(
        eln_import_id: int
) -> typing.Optional[str]:
    eln_import = eln_imports.ELNImport.query.filter_by(id=eln_import_id).first()
    if eln_import is None:
        return None
    return eln_import.signed_by


def _replace_references(
        data: typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]],
        eln_source_url: str,
        generate_sampledb_object_url: bool = True
) -> None:
    """
    Replace references to users and objects.

    :param data: the data to replace references in
    :param eln_source_url: the URL of the ELN the data is from
    :param generate_sampledb_object_url: whether object URLs should be generated in the SampleDB pattern
    """
    if isinstance(data, list):
        for item in data:
            _replace_references(item, eln_source_url)
    elif isinstance(data, dict):
        if '_type' in data:
            if data['_type'] in ['user', 'sample', 'measurement', 'object_reference']:
                if 'eln_source_url' in data:
                    del data['eln_source_url']
                if 'eln_object_url' in data:
                    del data['eln_object_url']
                if 'eln_user_url' in data:
                    del data['eln_user_url']
                if 'component_uuid' not in data:
                    data['eln_source_url'] = eln_source_url
                    if data['_type'] == 'user':
                        if generate_sampledb_object_url and 'user_id' in data and isinstance(data['user_id'], int):
                            data['eln_user_url'] = eln_source_url + ('' if eln_source_url.endswith('/') else '/') + relative_url_for('frontend.user_profile', user_id=data['user_id'])
                        else:
                            data['eln_user_url'] = None
                    else:
                        if generate_sampledb_object_url and 'object_id' in data and isinstance(data['object_id'], int):
                            data['eln_object_url'] = eln_source_url + ('' if eln_source_url.endswith('/') else '/') + relative_url_for('frontend.object', object_id=data['object_id'])
                        else:
                            data['eln_object_url'] = None
        else:
            for value in data.values():
                _replace_references(value, eln_source_url)


def _remove_template_ids(
        schema: typing.Dict[str, typing.Any]
) -> None:
    """
    Remove references to template actions from a schema.

    :param schema: the schema to remove template action references from
    """
    if schema['type'] == 'array':
        _remove_template_ids(schema['items'])
    elif schema['type'] == 'object':
        if 'template' in schema:
            del schema['template']
        for property_schema in schema['properties'].values():
            _remove_template_ids(property_schema)


def _convert_property_value_to_id_schema_and_data(
        property_value: typing.Dict[str, typing.Any]
) -> typing.Tuple[str, typing.Dict[str, typing.Any], typing.Dict[str, typing.Any], str]:
    property_id = property_value['propertyID']
    # bool
    if isinstance(property_value['value'], bool):
        property_schema = {
            "type": "bool",
            "title": {
                "en": property_value.get('name', property_value['propertyID'])
            }
        }
        property_data = {
            "_type": "bool",
            "value": property_value['value']
        }
        text_representation = json.dumps(property_value['value'])
        return property_id, property_schema, property_data, text_representation
    # datetime
    if isinstance(property_value['value'], str):
        try:
            parsed_datetime = datetime.datetime.fromisoformat(property_value['value'])
        except ValueError:
            pass
        else:
            if parsed_datetime.tzinfo is not None:
                parsed_datetime = parsed_datetime.astimezone(datetime.timezone.utc)
            utc_datetime = parsed_datetime.replace(tzinfo=None)
            property_schema = {
                "type": "datetime",
                "title": {
                    "en": property_value.get('name', property_value['propertyID'])
                }
            }
            property_data = {
                "_type": "datetime",
                "utc_datetime": utc_datetime.strftime('%Y-%m-%d %H:%M:%S')
            }
            text_representation = property_value['value']
            return property_id, property_schema, property_data, text_representation
    # quantity
    if isinstance(property_value['value'], (int, float)):
        units = None
        if 'unitCode' in property_value:
            units = get_unit_for_un_cefact_code(property_value['unitCode'])
        if units is None and 'unitText' in property_value:
            try:
                pint_units = ureg.parse_units(property_value['unitText'])
            except Exception:
                pass
            else:
                units = str(pint_units)
        if units is None:
            units = "1"
        property_schema = {
            "type": "quantity",
            "title": {
                "en": property_value.get('name', property_value['propertyID'])
            },
            "units": units
        }
        property_data = {
            "_type": "quantity",
            "magnitude": property_value['value'],
            "units": units
        }
        if 'unitText' in property_value:
            text_representation = f"{property_value['value']} {property_value['unitText']}"
        else:
            text_representation = f"{property_value['value']}"
        return property_id, property_schema, property_data, text_representation
    # text as fallback
    text_representation = str(property_value['value'])
    property_schema = {
        "type": "text",
        "title": {
            "en": property_value.get('name', property_value['propertyID'])
        }
    }
    property_data = {
        "_type": "text",
        "text": {
            'en': text_representation
        }
    }
    if '\n' in text_representation:
        property_schema['multiline'] = True  # type: ignore
    return property_id, property_schema, property_data, text_representation


def _sanitize_keys(
        keys: typing.Sequence[str],
) -> typing.Union[typing.Dict[int, int], typing.Dict[int, str]]:
    keys_with_original_indices = sorted(zip(keys, range(len(keys))))
    used_indices: typing.List[typing.Tuple[int, int]] = []
    for key, original_index in keys_with_original_indices:
        try:
            index = int(key)
        except ValueError:
            break
        if index in dict(used_indices):
            break
        used_indices.append((index, original_index))
    else:
        used_indices.sort()
        offset = 0
        if used_indices[0][0] == 1:
            offset = 1
        if (used_indices[0][0] == offset) and (used_indices[-1][0] - offset == len(used_indices) - 1):
            return {
                original_index: int(key) - offset
                for key, original_index in keys_with_original_indices
            }
    sanitized_keys: typing.Dict[int, str] = {}
    for key, original_index in keys_with_original_indices:
        sanitized_key = ''.join(
            c
            for c in key.lower()
            if c in string.ascii_letters + string.digits + '_'
        )
        sanitized_key = sanitized_key.strip('_')
        while '__' in sanitized_key:
            sanitized_key = sanitized_key.replace('__', '_')
        if sanitized_key and sanitized_key[0] in string.ascii_letters and sanitized_key not in sanitized_keys.values():
            sanitized_keys[original_index] = sanitized_key
        else:
            if sanitized_key:
                sanitized_key = '_' + sanitized_key
            if 'property' + sanitized_key not in sanitized_keys.values():
                sanitized_keys[original_index] = 'property' + sanitized_key
            else:
                for i in itertools.count(2):
                    if f'property{i}' + sanitized_key not in sanitized_keys.values():
                        sanitized_keys[original_index] = f'property{i}' + sanitized_key
                        break
    return sanitized_keys


PropertyValue = typing.Dict[str, typing.Any]  # TODO: TypedDict?
PropertyPath = typing.Tuple[typing.Union[int, str], ...]
FlattenedPropertyValueTree = typing.Dict[typing.Sequence[typing.Union[int, str]], PropertyValue]


def _map_property_values_to_paths(property_values: typing.Sequence[PropertyValue]) -> FlattenedPropertyValueTree:
    if not property_values:
        return {}
    property_paths_and_values = []
    for property_value in property_values:
        property_id = property_value['propertyID']
        property_path: PropertyPath = tuple(property_id.split('.'))
        property_paths_and_values.append((property_path, property_value))
    max_depth = max(
        len(property_path)
        for property_path, _ in property_paths_and_values
    )
    for depth in range(max_depth):
        property_paths_and_values_by_prefix: typing.Dict[PropertyPath, typing.List[typing.Tuple[PropertyPath, PropertyValue]]] = {}
        sanitized_property_paths_and_values = []
        for property_path, property_value in property_paths_and_values:
            if len(property_path) <= depth:
                sanitized_property_paths_and_values.append((property_path, property_value))
                continue
            property_path_prefix = property_path[:depth]
            if property_path_prefix not in property_paths_and_values_by_prefix:
                property_paths_and_values_by_prefix[property_path_prefix] = []
            property_paths_and_values_by_prefix[property_path_prefix].append((property_path, property_value))
        for property_path_prefix, property_paths_and_values_for_prefix in property_paths_and_values_by_prefix.items():
            property_paths_and_values_by_path_element: typing.Dict[str, typing.Tuple[typing.List[typing.Tuple[PropertyPath, PropertyValue]], typing.List[typing.Tuple[PropertyPath, PropertyValue]]]] = {}
            for property_path, property_value in property_paths_and_values_for_prefix:
                path_element = typing.cast(str, property_path[depth])
                if path_element not in property_paths_and_values_by_path_element:
                    property_paths_and_values_by_path_element[path_element] = ([], [])
                if len(property_path) == depth + 1:
                    property_paths_and_values_by_path_element[path_element][0].append((property_path, property_value))
                else:
                    property_paths_and_values_by_path_element[path_element][1].append((property_path, property_value))
            unsanitized_path_elements = []
            property_paths_and_values_by_path_element_index: typing.List[typing.List[typing.Tuple[PropertyPath, PropertyValue]]] = []
            for path_element, property_path_and_values_for_path_element in property_paths_and_values_by_path_element.items():
                unsanitized_path_elements += [path_element] * len(property_path_and_values_for_path_element[0])
                property_paths_and_values_by_path_element_index.extend(
                    [(property_path, property_value)]
                    for property_path, property_value in property_path_and_values_for_path_element[0]
                )
                if property_path_and_values_for_path_element[1]:
                    unsanitized_path_elements.append(path_element)
                    property_paths_and_values_by_path_element_index.append(property_path_and_values_for_path_element[1])
            sanitized_path_elements_by_index = _sanitize_keys(unsanitized_path_elements)
            for path_element_index, sanitized_path_element in sanitized_path_elements_by_index.items():
                for property_path, property_value in property_paths_and_values_by_path_element_index[path_element_index]:
                    sanitized_property_paths_and_values.append((
                        property_path_prefix + (sanitized_path_element,) + property_path[depth + 1:],
                        property_value
                    ))
        property_paths_and_values = sanitized_property_paths_and_values
    return {
        property_path: property_value
        for property_path, property_value in property_paths_and_values
    }


def _convert_property_value_trees_to_schema_and_data(
        flattened_property_value_tree: FlattenedPropertyValueTree,
        is_root: bool = False
) -> typing.Tuple[typing.Dict[str, typing.Any], typing.Any, str, typing.Sequence[str]]:
    if () in flattened_property_value_tree:
        _, schema, data, text = _convert_property_value_to_id_schema_and_data(flattened_property_value_tree[()])
        property_value = flattened_property_value_tree[()]
        full_title = property_value['propertyID']
        titles = full_title.split('.')
        titles = [
            title.strip() for title in titles
        ]
        if titles:
            schema['title'] = {
                'en': titles[-1] or 'Property'
            }
        return schema, data, text, titles
    first_path_elements = [
        property_path[0]
        for property_path in flattened_property_value_tree
    ]
    if first_path_elements and not is_root and all(isinstance(first_path_element, int) for first_path_element in first_path_elements):
        items = [
            _convert_property_value_trees_to_schema_and_data(
                {
                    property_path[1:]: property_value
                    for property_path, property_value in flattened_property_value_tree.items()
                    if property_path[0] == index
                }
            )
            for index in sorted(first_path_elements)
        ]
        if not items:
            return {
                'title': {
                    'en': 'Array'
                },
                'type': 'array',
                'items': {
                    'title': 'Item',
                    'type': 'text'
                }
            }, [], '', ['']
        # check all schemas are the same / compatible (for quantities) and otherwise fall back to use text
        first_item_schema = copy.deepcopy(items[0][0])
        if all(item[0] == first_item_schema for item in items):
            compatible_item_schema = first_item_schema
        else:
            item_schema_types = {item[0].get('type') for item in items}
            if item_schema_types == {'quantity'}:
                used_units = {
                    item[0]['units'] for item in items
                }
                if len(used_units) == 1:
                    compatible_item_schema = first_item_schema
                else:
                    compatible_item_schema = None
            elif item_schema_types == {'object'}:
                compatible_item_schema = None  # TODO: implement this
            elif item_schema_types == {'array'}:
                compatible_item_schema = None  # TODO: implement this
            elif len(item_schema_types) == 1:
                # while other schema types generally support incompatible restrictions, e.g. limiting object_references to specific action types, these features are not used here
                compatible_item_schema = first_item_schema
            else:
                compatible_item_schema = None
        if compatible_item_schema is not None:
            item_schema = compatible_item_schema
        else:
            item_schema = {
                'type': 'text',
                'multiline': any('\n' in text for _, data, text, _ in items)
            }
        item_schema['title'] = {
            'en': 'Item'
        }
        titles = items[0][3][:-1]
        if not titles:
            titles = ['Array']
        return {
            'title': {
                'en': titles[-1]
            },
            'type': 'array',
            'items': item_schema
        }, [
            data if compatible_item_schema is not None else {
                '_type': 'text',
                'text': {
                    'en': text
                }
            }
            for _, data, text, _ in items
        ], '\n'.join(text for _, _, text, _ in items), titles
    else:
        properties = {
            property_name: _convert_property_value_trees_to_schema_and_data(
                {
                    property_path[1:]: property_value
                    for property_path, property_value in flattened_property_value_tree.items()
                    if property_path[:1] == (property_name,)
                }
            )
            for property_name in first_path_elements
        }
        if properties:
            titles = next(iter(properties.values()))[3][:-1]
        else:
            titles = []
        if not titles:
            titles = ['Object']
        return {
            'title': {
                'en': titles[-1]
            },
            'type': 'object',
            'properties': {
                path_element: schema_data_titles[0]
                for path_element, schema_data_titles in properties.items()
            },
            'propertyOrder': [],
            'required': []
        }, {
            path_element: schema_data_titles[1]
            for path_element, schema_data_titles in properties.items()
        }, '', titles  # TODO: implement text


def _insert_property_into_schema_and_data(
        schema: typing.Dict[str, typing.Any],
        data: typing.Dict[str, typing.Any],
        property_name: str,
        property_title: str,
        property_value: typing.Dict[str, typing.Any],
        property_required: bool,
        property_order_position: int
) -> None:
    if property_required:
        if property_name not in schema['required']:
            schema['required'].append(property_name)
    if property_name in schema['propertyOrder']:
        schema['propertyOrder'].remove(property_name)
    schema['propertyOrder'].insert(property_order_position, property_name)
    # if property is already in data with the correct text and schema, there is no need to duplicate it
    if property_name in data and data[property_name] == property_value and property_name in schema['properties'] and schema['properties'][property_name].get('type') == property_value.get('_type') and not schema['properties'][property_name].get('multiline') and not schema['properties']['name'].get('markdown'):
        return
    if property_name in schema['properties']:
        for i in itertools.count(1):
            alternative_property_name = f'property{i if i > 1 else ""}_{property_name}'
            if alternative_property_name not in schema['properties']:
                schema['properties'][alternative_property_name] = schema['properties'][property_name]
                data[alternative_property_name] = data[property_name]
                break
    schema['properties'][property_name] = {
        'type': property_value.get('_type'),
        'title': {
            'en': property_title
        }
    }
    data[property_name] = property_value


def _convert_property_values_to_data_and_schema(
        property_values: typing.Sequence[PropertyValue],
        name: str,
        description: str,
        description_is_markdown: bool,
        tags: typing.Sequence[str]
) -> typing.Tuple[typing.Dict[str, typing.Any], typing.Dict[str, typing.Any]]:
    flattened_property_value_tree = _map_property_values_to_paths(property_values=property_values)
    schema, data, _, _ = _convert_property_value_trees_to_schema_and_data(
        flattened_property_value_tree=flattened_property_value_tree,
        is_root=True
    )
    _insert_property_into_schema_and_data(
        schema=schema,
        data=data,
        property_name='name',
        property_title='Name',
        property_value={'_type': 'text', 'text': {'en': name}},
        property_required=True,
        property_order_position=0
    )
    if description:
        _insert_property_into_schema_and_data(
            schema=schema,
            data=data,
            property_name='description',
            property_title='Description',
            property_value={'_type': 'text', 'text': {'en': description}},
            property_required=False,
            property_order_position=1
        )
        if description_is_markdown:
            data['description']['is_markdown'] = True
            schema['properties']['description']['markdown'] = True
    if tags:
        _insert_property_into_schema_and_data(
            schema=schema,
            data=data,
            property_name='tags',
            property_title='Tags',
            property_value={'_type': 'tags', 'tags': tags},
            property_required=False,
            property_order_position=len(schema['propertyOrder'])
        )
    return data, schema



# coding: utf-8
"""
Authentication functions for the SampleDB RESTful API.
"""
import typing

import flask

from flask_httpauth import HTTPBasicAuth, HTTPTokenAuth, MultiAuth

from ...logic.authentication import login, login_via_api_token, login_via_api_access_token, login_via_api_refresh_token, login_via_oidc_access_token, get_active_two_factor_authentication_methods, generate_api_access_token, refresh_api_access_token
from ...logic.users import User
from ...models import Permissions
from ...utils import object_permissions_required as object_permissions_required_generic
from ..utils import Resource, ResponseData

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'

http_basic_auth = HTTPBasicAuth()
http_api_token_or_access_token_auth = HTTPTokenAuth(scheme='Bearer')
http_api_token_or_refresh_token_auth = HTTPTokenAuth(scheme='Bearer')
multi_auth = MultiAuth(http_basic_auth, http_api_token_or_access_token_auth)
multi_auth_for_access_tokens = MultiAuth(http_basic_auth, http_api_token_or_refresh_token_auth)


@http_api_token_or_access_token_auth.verify_token
def verify_api_token_or_access_token(api_token: typing.Optional[str]) -> typing.Optional[User]:
    if not api_token:
        return None

    # login_via_oidc_access_token should be the last method to be called,
    # otherwise, if OIDC is configured, even the login using other types of
    # API tokens would be blocked on an HTTP request.
    # Additionally, this avoids unnecessarily sharing valid API tokens with
    # the provider, even if the provider is trusted.
    user = login_via_api_token(api_token)
    if user is None:
        user = login_via_api_access_token(api_token)
    if user is None:
        user = login_via_oidc_access_token(api_token)

    if user is None or not user.is_active:
        return None
    flask.g.user = user
    return user


@http_api_token_or_refresh_token_auth.verify_token
def verify_api_token_or_refresh_token(api_token: typing.Optional[str]) -> typing.Optional[User]:
    if not api_token:
        return None
    user = login_via_api_token(api_token)
    if user is None:
        user = login_via_api_refresh_token(api_token)
        if user is not None:
            flask.g.api_refresh_token = api_token
    if user is None or not user.is_active:
        return None
    flask.g.user = user
    return user


@http_basic_auth.verify_password
def verify_password(username: str, password: str) -> typing.Optional[User]:
    if not username:
        return None
    user = login(username, password)
    if user is None or not user.is_active:
        return None
    two_factor_authentication_methods = get_active_two_factor_authentication_methods(user.id)
    if two_factor_authentication_methods:
        # two-factor authentication is not supported for the HTTP API
        return None
    flask.g.user = user
    return user


def object_permissions_required(permissions: Permissions) -> typing.Callable[[typing.Any], typing.Any]:
    """
    Only allow access to a route it the user has the required permissions.

    Wrapper around the more generic sampledb.utils.object_permissions_required
    for use with the http_basic_auth object from this module.

    :param permissions: the required object permissions
    """
    return object_permissions_required_generic(
        required_object_permissions=permissions,
        auth_extension=multi_auth,
        user_id_callable=lambda: typing.cast(int, flask.g.user.id),
        may_enable_anonymous_users=False
    )


class AccessTokens(Resource):
    @multi_auth_for_access_tokens.login_required
    def post(self) -> ResponseData:
        refresh_token = getattr(flask.g, 'api_refresh_token', None)
        if refresh_token:
            result = refresh_api_access_token(refresh_token)
            if result is None:
                return {
                    "message": "failed to refresh API token"
                }, 400
            return result, 201
        request_json = flask.request.get_json(force=True, silent=True)
        if not isinstance(request_json, dict):
            return {
                "message": "JSON body should contain object"
            }, 400
        if set(request_json.keys()) != {'description'}:
            return {
                "message": "JSON body should only contain description"
            }, 400
        description = request_json.get('description')
        if not isinstance(description, str):
            return {
                "message": "description must be a string"
            }, 400
        return generate_api_access_token(user_id=flask.g.user.id, description=description), 201



# coding: utf-8
"""
Add ASSIGN_LOCATION enum value to UserLogEntryType enum.
"""

import flask_sqlalchemy

from .utils import enum_value_migration


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    return enum_value_migration('userlogentrytype', 'ASSIGN_LOCATION')



# coding: utf-8
"""

"""

import flask

from . import frontend
from ..utils import FlaskResponseT

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


@frontend.route('/')
def index() -> FlaskResponseT:
    return flask.render_template(
        'index.html'
    )



# coding: utf-8
"""
Authentication functions for the SampleDB federation RESTful API.
"""
import typing

import flask

from flask_httpauth import HTTPTokenAuth

from ...logic.component_authentication import login_via_component_token
from ...logic.components import Component

http_token_auth = HTTPTokenAuth(scheme='Bearer')


@http_token_auth.verify_token
def verify_token(component_token: typing.Optional[str]) -> typing.Optional[Component]:
    if not component_token:
        return None
    component = login_via_component_token(component_token)
    if component is None:
        return None
    flask.g.component = component
    return component



# coding: utf-8
"""
Replace string columns with JSON columns in groups table.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    column_names = db.session.execute(db.text("""
        SELECT column_name, data_type
        FROM information_schema.columns
        WHERE table_name = 'groups'
    """)).fetchall()
    if ('name', 'json') in column_names and ('description', 'json') in column_names:
        return False

        # Perform migration
    existing_data = [
        group_data
        for group_data in db.session.execute(db.text("""
               SELECT id, name, description
               FROM groups
           """)).fetchall()
    ]

    db.session.execute(db.text("""
           ALTER TABLE groups
           DROP COLUMN name;
       """))
    db.session.execute(db.text("""
           ALTER TABLE groups
           DROP COLUMN description;
       """))
    db.session.execute(db.text("""
           ALTER TABLE groups
           ADD name JSON NOT NULL DEFAULT '{}'::json;
       """))
    db.session.execute(db.text("""
           ALTER TABLE groups
           ADD description JSON NOT NULL DEFAULT '{}'::json;
       """))
    for id, name, description in existing_data:
        group_data = {
            'id': id,
            'name': name,
            'description': description
        }
        db.session.execute(db.text("""
               UPDATE groups
               SET name = json_build_object('en', :name), description = json_build_object('en', :description)
               WHERE id = :id;
           """), params=group_data)
    return True



import typing

from .. import actions
from ... import logic
from ..errors import InvalidNumberError, InvalidTemplateIDError, RecursiveTemplateError, ValidationError
from ...models import ActionType, Permissions
from .utils import schema_iter

# keys and properties that only can be used in root objects

SKIPPED_TEMPLATE_KEYS = {
    'displayProperties',
    'batch',
    'batch_name_format',
    'notebookTemplates',
}
SKIPPED_TEMPLATE_PROPERTY_NAMES = {
    'name',
    'tags',
    'hazards'
}


def process_template_action_schema(
        template_action_schema: typing.Dict[str, typing.Any]
) -> typing.Dict[str, typing.Any]:
    processed_schema: typing.Dict[str, typing.Any] = {}
    for key in template_action_schema:
        if key == 'properties':
            processed_schema['properties'] = {
                property_name: property_value
                for property_name, property_value in template_action_schema['properties'].items()
                if property_name not in SKIPPED_TEMPLATE_PROPERTY_NAMES
            }
        elif key == 'required':
            processed_schema['required'] = [
                property_name
                for property_name in template_action_schema['required']
                if property_name not in SKIPPED_TEMPLATE_PROPERTY_NAMES
            ]
        elif key == 'propertyOrder':
            processed_schema['propertyOrder'] = [
                property_name
                for property_name in template_action_schema['propertyOrder']
                if property_name not in SKIPPED_TEMPLATE_PROPERTY_NAMES
            ]
        elif key == 'show_more':
            processed_schema['show_more'] = [
                property_name
                for property_name in template_action_schema['show_more']
                if property_name not in SKIPPED_TEMPLATE_PROPERTY_NAMES
            ]
        elif key not in SKIPPED_TEMPLATE_KEYS:
            processed_schema[key] = template_action_schema[key]
    return processed_schema


def substitute_templates(
        schema: typing.Dict[str, typing.Any],
        invalid_template_action_ids: typing.Sequence[int] = ()
) -> None:
    if 'template' in schema.keys():
        if type(schema['template']) is dict and 'action_id' in schema['template'] and 'component_uuid' in schema['template']\
                and type(schema['template']['action_id']) is int and type(schema['template']['component_uuid']) is str:
            return
        if type(schema['template']) is not int:
            raise InvalidNumberError()
        if schema['template'] in invalid_template_action_ids:
            raise RecursiveTemplateError()
        for key in ['properties', 'required', 'propertyOrder']:
            if key in schema:
                del schema[key]
        template_action = actions.get_action(schema['template'])
        if template_action.schema is None or template_action.type is None or not (template_action.type.is_template or template_action.type.fed_id == ActionType.TEMPLATE):
            raise InvalidTemplateIDError()
        template_schema = template_action.schema
        template_schema = process_template_action_schema(template_schema)
        if schema.get('title'):
            del template_schema['title']
        if 'show_more' in schema and 'show_more' in template_schema:
            del template_schema['show_more']
        if 'note' in schema and 'note' in template_schema:
            del template_schema['note']
        schema.update(template_schema)


def reverse_substitute_templates(schema: typing.Dict[str, typing.Any]) -> None:
    root_schema = schema
    for _, schema in schema_iter(root_schema, filter_property_types={'object'}):
        if 'template' in schema:
            # ensure the action exists
            actions.check_action_exists(schema['template'])
            for key in list(schema.keys()):
                if key in ['title', 'type', 'template', 'may_copy', 'conditions']:
                    continue
                elif key in ['properties']:
                    schema[key] = {}
                elif key in ['propertyOrder', 'required']:
                    schema[key] = []
                else:
                    del schema[key]


def update_schema_using_template_action(
        schema: typing.Dict[str, typing.Any],
        template_action_id: int,
        template_action_schema: typing.Dict[str, typing.Any]
) -> typing.Dict[str, typing.Any]:
    root_schema = schema
    for _, schema in schema_iter(root_schema, filter_property_types={'object'}):
        if schema.get('template') == template_action_id:
            if schema.get('title'):
                title = schema['title']
            else:
                title = None
            schema.update(template_action_schema)
            if title is not None:
                schema['title'] = title
    return root_schema


def find_invalid_template_paths(
        schema: typing.Dict[str, typing.Any],
        user_id: int
) -> typing.Sequence[typing.List[str]]:
    """
    Find all property paths to object schemas with an invalid template action.

    :param schema: the schema to check for invalid template paths
    :param user_id: the user ID to check permissions with
    :return: a list of property paths for invalid templates
    """
    invalid_template_paths = []
    if not isinstance(schema, dict):
        raise ValidationError('schema must be dict', [])
    for property_path, property_schema in schema_iter(schema, filter_property_types={'object'}):
        template_action_id = property_schema.get('template')
        if type(template_action_id) is int:
            template_action_permissions = logic.action_permissions.get_user_action_permissions(
                action_id=template_action_id,
                user_id=user_id
            )
            if Permissions.READ not in template_action_permissions:
                invalid_template_paths.append([
                    path_element or '[?]'
                    for path_element in property_path
                ])
    return invalid_template_paths


def find_used_template_ids(
        schema: typing.Dict[str, typing.Any]
) -> typing.Set[int]:
    """
    Find all template action IDs used by the schema.

    :param schema: the schema to search for template objects
    :return: a set of template action IDs
    """
    template_action_ids = set()
    for _property_path, property_schema in schema_iter(schema, filter_property_types={'object'}):
        template_action_id = property_schema.get('template')
        if type(template_action_id) is int:
            template_action_ids.add(template_action_id)
    return template_action_ids



import typing

import flask

from .. import errors
from ..components import add_component, get_component_by_uuid, Component, ComponentInfo, add_or_update_component_info
from .utils import _get_uuid, _get_str, _get_int, _get_bool
from ..components import validate_address, MAX_COMPONENT_ADDRESS_LENGTH, MAX_COMPONENT_NAME_LENGTH


def _get_or_create_component_id(
        component_uuid: str
) -> typing.Optional[int]:
    if component_uuid == flask.current_app.config['FEDERATION_UUID']:
        return None
    try:
        component = get_component_by_uuid(component_uuid)
    except errors.ComponentDoesNotExistError:
        component = add_component(uuid=component_uuid, description='', name=None, address=None)
    return component.id


class ComponentData(typing.TypedDict):
    uuid: str
    source_uuid: str
    name: typing.Optional[str]
    address: typing.Optional[str]
    discoverable: bool
    distance: int


def parse_component_info(
        component_data: typing.Dict[str, typing.Any],
        component: Component
) -> ComponentData:
    address = _get_str(component_data.get('address'))
    if address is not None:
        try:
            address = validate_address(
                address,
                max_length=MAX_COMPONENT_ADDRESS_LENGTH,
                allow_http=flask.current_app.config['ALLOW_HTTP']
            )
        except Exception:
            address = None
    name = _get_str(component_data.get('name'))
    if name is not None:
        if len(name) > MAX_COMPONENT_NAME_LENGTH:
            name = name[:MAX_COMPONENT_NAME_LENGTH]
    return ComponentData(
        uuid=_get_uuid(component_data.get('uuid')),
        source_uuid=_get_uuid(component_data.get('source_uuid')),
        name=name,
        address=address,
        discoverable=_get_bool(component_data.get('discoverable'), mandatory=True),
        distance=_get_int(component_data.get('distance'), mandatory=True)
    )


def import_component_info(
        component_data: ComponentData,
        component: Component
) -> ComponentInfo:
    return add_or_update_component_info(
        uuid=component_data['uuid'],
        source_uuid=component_data['source_uuid'],
        name=component_data['name'],
        address=component_data['address'],
        discoverable=component_data['discoverable'],
        distance=component_data['distance']
    )


def parse_import_component_info(
        component_data: typing.Dict[str, typing.Any],
        component: Component
) -> ComponentInfo:
    return import_component_info(parse_component_info(component_data, component), component)



# coding: utf-8
"""
RESTful API for SampleDB
"""
import typing

from .authentication import multi_auth
from ..utils import Resource, ResponseData
from ...logic.action_types import get_action_types, get_action_type
from ...logic import errors, utils, action_types

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def action_type_to_json(action_type: action_types.ActionType) -> typing.Dict[str, typing.Any]:
    return {
        'type_id': action_type.id,
        'name': utils.get_translated_text(
            action_type.name,
            language_code='en',
            default='Unnamed Action Type'
        ),
        'object_name': utils.get_translated_text(
            action_type.object_name,
            language_code='en',
            default='Object'
        ),
        'admin_only': action_type.admin_only
    }


class ActionType(Resource):
    @multi_auth.login_required
    def get(self, type_id: int) -> ResponseData:
        try:
            action_type = get_action_type(
                action_type_id=type_id
            )
        except errors.ActionTypeDoesNotExistError:
            return {
                "message": f"action type {type_id} does not exist"
            }, 404
        return action_type_to_json(action_type)


class ActionTypes(Resource):
    @multi_auth.login_required
    def get(self) -> ResponseData:
        action_types = get_action_types()
        return [
            action_type_to_json(action_type)
            for action_type in action_types
        ]



# coding: utf-8
"""
Set WRITE permissions for all users for all existing locations if there are no
location permissions yet.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    existing_permissions = db.session.execute(db.text("""
        SELECT 1 FROM all_user_location_permissions
        UNION
        SELECT 1 FROM user_location_permissions
        UNION
        SELECT 1 FROM group_location_permissions
        UNION
        SELECT 1 FROM project_location_permissions
    """)).fetchall()
    if existing_permissions:
        return False

    # Perform migration
    existing_locations = db.session.execute(db.text("""
        SELECT id
        FROM locations
    """)).fetchall()
    for existing_location in existing_locations:
        db.session.execute(db.text("""
           INSERT INTO all_user_location_permissions
           (location_id, permissions)
           VALUES
           (:location_id, 'WRITE')
       """), {'location_id': existing_location[0]})
    return True



# coding: utf-8
"""

"""

import importlib
import sys
import typing

from ... import db


def should_skip_by_index(db: typing.Any, index: int) -> bool:
    """
    Returns whether or not a migration should be skipped due to its index.

    :param db: the database
    :param index: the migration index to check
    :return: whether or not the migration should be skipped
    """
    # migration 0 creates the migration_index table, so it cannot be skipped
    # by index as the table might not exist yet
    if index == 0:
        return False

    return db.session.execute(
        db.text("""
        SELECT migration_index
        FROM migration_index
        WHERE migration_index >= :index
        """),
        {'index': index}
    ).first() is not None


def update_migration_index(db: typing.Any, index: int) -> None:
    """
    Set the database's migration index to the given index.

    This function also ensures that the migration index can only ever be
    increased and fails with an assertion error otherwise.

    :param db: the database
    :param index: the new migration index
    """
    # migration 0 creates the migration_index table and sets it to 0
    if index == 0:
        return

    assert db.session.execute(
        db.text("""
        UPDATE migration_index
        SET migration_index = :index
        WHERE migration_index < :index
        """),
        {"index": index}
    ).rowcount == 1


def get_migrations() -> typing.List[typing.Tuple[int, str, typing.Callable[[typing.Any], bool]]]:
    """
    Finds migrations and returns them sorted by their index.

    :return: A list of migrations, with each migration as a tuple of its
        index, name and function.
    """
    migration_module_names = [
        "create_migration_index",
        "actions_add_user_id",
        "object_log_entry_type_add_assign_location",
        "user_log_entry_type_add_assign_location",
        "user_log_entry_type_add_create_location",
        "user_log_entry_type_add_update_location",
        "object_location_assignments_add_responsible_user_id",
        "create_user_object_permissions_by_all_view",
        "files_add_data",
        "user_log_entry_type_add_link_publication",
        "object_log_entry_type_add_link_publication",
        "object_location_assignments_add_confirmed",
        "user_add_is_readonly",
        "user_add_is_hidden",
        "authentication_type_add_api_token",
        "actions_add_description_as_html",
        "instruments_add_description_as_html",
        "user_add_orcid",
        "user_log_entry_type_add_create_instrument_log_entry",
        "notification_type_add_instrument_log_entry_created",
        "instruments_add_users_can_create_log_entries",
        "instruments_add_users_can_view_log_entries",
        "instruments_add_notes",
        "instruments_add_notes_as_html",
        "user_add_affiliation",
        "instruments_add_create_log_entry_default",
        "actions_add_is_hidden",
        "publications_add_object_name",
        "instruments_add_is_hidden",
        "notification_type_add_referenced_by_object_metadata",
        "action_type_create_default_action_types",
        "action_replace_type_with_type_id",
        "object_log_entry_type_reference_object_in_metadata",
        "user_add_is_active",
        "notification_type_add_instrument_log_entry_edited",
        "instrument_log_entries_add_versioning",
        "object_log_entry_type_export_to_dataverse",
        "actions_replace_description_as_html_with_description_is_markdown",
        "instruments_replace_description_as_html_with_description_is_markdown",
        "instruments_replace_notes_as_html_with_notes_is_markdown",
        "instruments_add_short_description",
        "actions_add_short_description",
        "markdown_to_html_cache_recreate",
        "action_type_add_enable_project_link",
        "object_log_entry_type_add_link_project",
        "object_log_entry_type_add_unlink_project",
        "instrument_log_entry_versions_add_content_is_markdown",
        "public_actions_add_common_actions",
        "instrument_log_entry_versions_add_event_utc_datetime",
        "languages_create_default_languages",
        "actions_move_to_action_translations",
        "action_type_move_to_translation",
        "action_type_translations_create_default_translations",
        "instruments_move_to_translations",
        "locations_replace_string_columns_with_JSON",
        "groups_replace_string_columns_with_JSON",
        "projects_replace_string_columns_with_JSON",
        "location_assignments_replace_string_columns_with_JSON",
        "user_add_role",
        "languages_add_enabled_for_user_interface",
        "action_type_drop_description",
        "actions_drop_description_as_html",
        "actions_drop_short_description_as_html",
        "instruments_drop_description_as_html",
        "instruments_drop_short_description_as_html",
        "instruments_drop_notes_as_html",
        "user_add_extra_fields",
        "files_add_binary_data",
        "action_type_add_disable_create_objects",
        "action_type_add_template_attribute",
        "action_type_add_schema_template",
        "objects_current_add_fed_key",
        "objects_current_update_not_null_constraints",
        "objects_previous_add_fed_key",
        "objects_previous_update_not_null_constraints",
        "instruments_add_fed_key",
        "instrument_translations_update_not_null_constraints",
        "users_add_fed_key",
        "user_type_add_federation_user",
        "actions_add_fed_key",
        "actions_update_not_null_constraints",
        "action_types_add_fed_key",
        "locations_add_fed_key",
        "locations_update_not_null_constraints",
        "object_location_assignments_add_fed_key",
        "object_location_assignments_update_not_null_constraints",
        "files_add_fed_key",
        "files_update_not_null_constraints",
        "comments_add_fed_key",
        "comments_update_not_null_constraints",
        "users_update_not_null_constraints",
        "action_translations_update_not_null_constraints",
        "markdown_images_add_fed_key",
        "create_objects_current_name_cache",
        "create_objects_current_tags_cache",
        "drop_public_actions",
        "drop_public_objects",
        "replace_user_object_permissions_by_all_view",
        "drop_default_public_permissions",
        "setup_default_location_permissions",
        "file_log_add_edit_url_type",
        "replace_user_object_permissions_by_all_view2",
        "languages_add_datetime_format_moment_output",
        "replace_user_object_permissions_by_all_view3",
        "fed_user_aliases_add_use_real_name_column",
        "fed_user_aliases_add_use_real_email_column",
        "fed_user_aliases_add_use_real_orcid_column",
        "fed_user_aliases_add_use_real_affiliation_column",
        "fed_user_aliases_add_use_real_role_column",
        "fed_user_aliases_add_null_if_use_real_data_check",
        "action_type_add_scicat_export_type",
        "object_location_assignments_add_declined",
        "fed_user_aliases_add_last_modified",
        "components_add_last_sync_timestamp",
        "users_add_last_modified",
        "location_types_create_default_location_types",
        "locations_add_type_id",
        "location_log_create_initial_log_entries",
        "notification_type_add_responsibility_assignment_declined",
        "action_type_add_order_index",
        "actions_add_admin_only",
        "actions_add_disable_create_objects",
        "users_add_last_modified_by_id",
        "locations_add_is_hidden",
        "background_tasks_add_result_expiration_date_column",
        "dataverse_exports_add_status",
        "location_types_add_enable_instruments",
        "instruments_add_location_id",
        "files_fix_fed_unique_constraint",
        "object_location_assignments_drop_declined_default",
        "instruments_drop_notes",
        "object_location_assignments_drop_utc_datetime_not_null_constraint",
        "locations_drop_is_hidden_default",
        "languages_change_datetime_format_moment_output_type",
        "replace_user_object_permissions_by_all_view4",
        "fed_object_log_entries_add_user_id",
        "object_shares_add_user_id",
        "fed_object_log_entry_type_add_remote_object_import",
        "object_shares_add_import_status",
        "notification_type_add_remote_object_import_failed",
        "notification_type_add_remote_object_import_notes",
        "location_types_add_enable_capacities",
        "action_translations_add_not_null_constraints",
        "action_type_translations_add_not_null_constraints",
        "instrument_translations_add_not_null_constraints",
        "authentications_add_not_null_constraints",
        "component_authentications_add_not_null_constraints",
        "own_component_authentications_add_not_null_constraints",
        "download_service_job_files_add_not_null_constraints",
        "two_factor_authentication_methods_add_not_null_constraints",
        "markdown_images_add_not_null_constraint",
        "languages_add_not_null_constraints",
        "action_type_add_enable_instrument_link",
        "instruments_add_object_id",
        "instruments_add_show_linked_object_data",
        "authentication_type_add_api_access_token",
        "user_type_add_eln_import_user",
        "object_log_entry_type_add_import_from_eln_file",
        "objects_previous_add_eln_import",
        "objects_current_add_eln_import",
        "users_add_eln_import",
        "user_log_entry_type_add_import_from_eln_file",
        "components_add_discoverable",
        "actions_add_objects_readable_by_all_users_by_default",
        "files_add_data_not_null_constraint",
        "replace_timestamp_with_timestamptz",
        "authentication_type_add_fido2_passkey",
        "action_type_add_show_in_object_filters",
        "notification_type_add_automatic_user_federation",
        "topics_add_short_description",
        "topics_add_is_markdown",
        "actions_add_use_instrument_topics",
        "object_log_add_is_imported",
        "languages_add_date_format_moment_output",
        "files_add_preview_image",
        "group_invitations_add_revoked",
        "project_invitations_add_revoked",
        "instrument_log_file_attachments_generate_image_infos",
        "fed_identity_add_login_column",
        "components_add_fed_login_available_column",
        "authentication_type_add_federated_login",
        "locations_add_enable_object_assignments",
        "authentication_type_add_oidc",
        "notification_mode_for_types_add_not_null_constraint",
        'eln_imports_add_signed_by',
    ]

    migrations = []
    for migration_index, migration_module_name in enumerate(migration_module_names):
        try:
            migration_module = importlib.import_module('sampledb.models.migrations.' + migration_module_name)
            migration_function = migration_module.run
        except Exception:
            print(f'Failed to load migration #{migration_index} "{migration_module_name}".', file=sys.stderr)
            sys.exit(1)
        migrations.append((migration_index, migration_module_name, migration_function))
    return migrations


def table_has_column(table_name: str, column_name: str) -> bool:
    """
    Return whether a table has a column with a given name.

    :param table_name: the name of the table
    :param column_name: the name of the column to check for
    :return: whether the column exists
    """
    return bool(db.session.execute(
        db.text("""
            SELECT COUNT(*)
            FROM information_schema.columns
            WHERE table_name = :table_name AND column_name = :column_name
        """),
        params={
            'table_name': table_name,
            'column_name': column_name
        }
    ).scalar())


def column_is_nullable(table_name: str, column_name: str) -> bool:
    """
    Return whether a column may contain NULL values.

    :param table_name: the name of the table
    :param column_name: the name of the column to check
    :return: whether the column may contain NULL values
    """
    return bool(db.session.execute(
        db.text("""
            SELECT is_nullable
            FROM information_schema.columns
            WHERE table_name = :table_name AND column_name = :column_name
        """),
        params={
            'table_name': table_name,
            'column_name': column_name
        }
    ).scalar() == 'YES')


def enum_has_value(enum_name: str, value_name: str) -> bool:
    """
    Return whether an enum has a value with a given name.

    :param enum_name: the name of the enum
    :param value_name: the name of the value to check for
    :return: whether the value exists
    """
    return bool(db.session.execute(
        db.text("""
            SELECT COUNT(*)
            FROM pg_type
            JOIN pg_enum ON pg_enum.enumtypid = pg_type.oid
            WHERE pg_type.typname = :enum_name AND pg_enum.enumlabel = :value_name
        """),
        params={
            'enum_name': enum_name,
            'value_name': value_name
        }
    ).scalar())


def add_enum_value(enum_name: str, value_name: str) -> None:
    """
    Add a value to an enum.

    :param enum_name: the name of the enum
    :param value_name: the name of the value to add
    """
    # Use connection and run COMMIT as ALTER TYPE cannot run in a transaction (in PostgreSQL 11)
    engine = db.engine.execution_options(autocommit=False)
    with engine.connect() as connection:
        connection.execute(db.text("COMMIT"))
        connection.execute(db.text(f"""
            ALTER TYPE {enum_name}
            ADD VALUE '{value_name}'
        """))


def enum_value_migration(enum_name: str, value_name: str) -> bool:
    """
    Perform a migration for adding a value to an enum, if the value is missing.

    :param enum_name: the name of the enum
    :param value_name: the name of the value to add
    :return: whether the migration was performed
    """
    if enum_has_value(enum_name, value_name):
        return False
    add_enum_value(enum_name, value_name)
    return True


def timestamp_add_timezone_utc_migration(table_name: str, column_name: str) -> bool:
    """
    Perform a migration for converting a TIMESTAMP WITHOUT TIME ZONE column to
    type TIMESTAMP WITH TIME ZONE using UTC for the conversion.

    :param table_name: the name of the table
    :param column_name: the name of the column to convert
    :return: whether the migration was performed
    """
    if db.session.execute(
        db.text("""
            SELECT data_type
            FROM information_schema.columns
            WHERE table_name = :table_name AND column_name = :column_name
        """),
        params={
            'table_name': table_name,
            'column_name': column_name
        }
    ).scalar() == 'timestamp with time zone':
        return False
    db.session.execute(db.text(f"""
        ALTER TABLE {table_name}
        ALTER {column_name}
        TYPE TIMESTAMP WITH TIME ZONE
        USING {column_name} AT TIME ZONE 'UTC'
    """))
    return True



# coding: utf-8
"""

"""
import dataclasses
import datetime
import typing

import minisign

from .. import db
from .. models import minisign_keys


@dataclasses.dataclass(frozen=True)
class KeyPair:
    """
    This class provides an immutable wrapper around models.minisign_keys.KeyPair.
    """
    id: int
    sk_bytes: bytes
    pk_bytes: bytes
    utc_datetime_created: datetime.datetime

    @classmethod
    def from_database(cls, key_pair: minisign_keys.KeyPair) -> 'KeyPair':
        return KeyPair(
            id=key_pair.id,
            sk_bytes=key_pair.sk_bytes,
            pk_bytes=key_pair.pk_bytes,
            utc_datetime_created=key_pair.utc_datetime_created
        )


def get_current_key_pair() -> KeyPair:
    kp = db.session.query(minisign_keys.KeyPair).first()
    if kp is None:
        return new_key_pair()
    return KeyPair.from_database(kp)


def new_key_pair() -> KeyPair:
    key_pair = minisign.KeyPair.generate()
    kp = minisign_keys.KeyPair(sk_bytes=bytes(key_pair.secret_key), pk_bytes=bytes(key_pair.public_key))
    db.session.add(kp)
    db.session.commit()
    return KeyPair.from_database(kp)


def get_key_pairs() -> typing.List[KeyPair]:
    kps = db.session.query(minisign_keys.KeyPair).all()
    return [KeyPair.from_database(kp) for kp in kps]


def get_key_pair_by_id(id: int) -> typing.Optional[KeyPair]:
    kp = db.session.query(minisign_keys.KeyPair).filter(minisign_keys.KeyPair.id == id).first()
    return KeyPair.from_database(kp) if kp else None



# coding: utf-8
"""
Splits action types into action_types and action_type_translations
"""

import flask_sqlalchemy

from .utils import table_has_column
from ..languages import Language


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if not table_has_column('instruments', 'name'):
        return False

    # Perform migration
    existing_data = [
        instrument_data
        for instrument_data in db.session.execute(db.text("""
            SELECT id, name, description, short_description, notes
            FROM instruments
        """)).fetchall()
    ]

    db.session.execute(db.text("""
            ALTER TABLE instruments
            DROP COLUMN name,
            DROP COLUMN description,
            DROP COLUMN short_description,
            DROP COLUMN notes
        """))

    performed_migration = False
    for instrument_id, name, description, short_description, notes in existing_data:
        translation = {
            'instrument_id': instrument_id,
            'language_id': Language.ENGLISH,
            'name': name,
            'description': description,
            'short_description': short_description,
            'notes': notes
        }
        # Perform migration
        db.session.execute(db.text("""
                      INSERT INTO instrument_translations (instrument_id, language_id, name, description, short_description, notes)
                      VALUES (:instrument_id, :language_id, :name, :description, :short_description, :notes)
                  """), params=translation)
        performed_migration = True

    return performed_migration



# coding: utf-8
"""

"""
import dataclasses
import datetime
import typing

from .objects import get_object, check_object_exists
from .components import check_component_exists, Component
from .notifications import create_notification_for_a_failed_remote_object_import, create_notification_for_a_remote_object_import_with_notes
from . import errors, fed_logs
from ..models import Object
from .. import db, models


@dataclasses.dataclass(frozen=True)
class ObjectShare:
    """
    This class provides an immutable wrapper around models.shares.ObjectShare.
    """
    object_id: int
    component_id: int
    policy: typing.Dict[str, typing.Any]
    utc_datetime: datetime.datetime
    component: Component
    user_id: typing.Optional[int]
    import_status: typing.Optional[typing.Dict[str, typing.Any]] = None

    @classmethod
    def from_database(cls, object_share: models.ObjectShare) -> 'ObjectShare':
        return ObjectShare(
            object_id=object_share.object_id,
            component_id=object_share.component_id,
            policy=object_share.policy,
            utc_datetime=object_share.utc_datetime,
            component=Component.from_database(object_share.component),
            user_id=object_share.user_id,
            import_status=object_share.import_status
        )


def get_all_shares() -> typing.List[ObjectShare]:
    """
    Returns a list of all existing shared objects.

    :return: the list of shares
    """
    return [
        ObjectShare.from_database(object_share)
        for object_share in models.ObjectShare.query.all()
    ]


def get_shares_for_component(component_id: int) -> typing.List[ObjectShare]:
    """
    Returns a list of all objects shared with a component.

    :param: the component's ID
    :return: the list of shares
    """
    shares = models.ObjectShare.query.filter_by(component_id=component_id).all()
    if len(shares) == 0:
        check_component_exists(component_id)
    return [
        ObjectShare.from_database(object_share)
        for object_share in shares
    ]


def get_shares_for_object(object_id: int) -> typing.List[ObjectShare]:
    shares = models.ObjectShare.query.filter_by(object_id=object_id).all()
    if len(shares) == 0:
        check_object_exists(object_id)
    return [
        ObjectShare.from_database(object_share)
        for object_share in shares
    ]


def get_object_if_shared(object_id: int, component_id: int) -> Object:
    obj = get_object(object_id)
    share = models.ObjectShare.query.filter_by(object_id=object_id, component_id=component_id).all()
    if len(share) == 0:
        check_component_exists(component_id)
        raise errors.ObjectNotSharedError()
    return obj


def _get_mutable_share(object_id: int, component_id: int) -> models.ObjectShare:
    share: typing.Optional[models.ObjectShare] = models.ObjectShare.query.filter_by(object_id=object_id, component_id=component_id).first()
    if share is None:
        check_object_exists(object_id)
        check_component_exists(component_id)
        raise errors.ShareDoesNotExistError()
    return share


def get_share(object_id: int, component_id: int) -> ObjectShare:
    share = _get_mutable_share(object_id, component_id)
    return ObjectShare.from_database(share)


def add_object_share(
        object_id: int,
        component_id: int,
        policy: typing.Dict[str, typing.Any],
        user_id: typing.Optional[int] = None
) -> ObjectShare:
    check_object_exists(object_id)
    check_component_exists(component_id)
    share = models.ObjectShare.query.filter_by(object_id=object_id, component_id=component_id).first()
    if share is not None:
        raise errors.ShareAlreadyExistsError()
    share = models.ObjectShare(object_id=object_id, component_id=component_id, policy=policy, user_id=user_id)
    db.session.add(share)
    db.session.commit()
    fed_logs.share_object(object_id, component_id, user_id=user_id)
    return ObjectShare.from_database(share)


def update_object_share(
        object_id: int,
        component_id: int,
        policy: typing.Dict[str, typing.Any],
        user_id: typing.Optional[int] = None
) -> ObjectShare:
    share = _get_mutable_share(object_id, component_id)
    if share.policy != policy:
        share.policy = policy
        share.user_id = user_id
        db.session.add(share)
        db.session.commit()
        fed_logs.update_object_policy(object_id, component_id, user_id=user_id)
    return ObjectShare.from_database(share)


def set_object_share_import_status(
        object_id: int,
        component_id: int,
        import_status: typing.Dict[str, typing.Any]
) -> None:
    share = _get_mutable_share(object_id, component_id)
    if share.import_status != import_status:
        if share.user_id:
            if not import_status['success'] and (share.import_status is None or share.import_status['success']):
                create_notification_for_a_failed_remote_object_import(share.user_id, object_id, component_id)
            elif import_status['success'] and import_status['notes'] and (share.import_status is None or not share.import_status['success'] or share.import_status['notes'] != import_status['notes']):
                create_notification_for_a_remote_object_import_with_notes(share.user_id, object_id, component_id, import_status['notes'])
        share.import_status = import_status
        db.session.add(share)
        db.session.commit()
        fed_logs.remote_import_object(
            object_id=object_id,
            component_id=component_id,
            import_status=import_status
        )


class ObjectShareImportStatus(typing.TypedDict):
    success: bool
    notes: typing.List[str]
    utc_datetime: str
    object_id: typing.Optional[int]


def parse_object_share_import_status(
        import_status: typing.Dict[str, typing.Any]
) -> typing.Optional[ObjectShareImportStatus]:
    if not isinstance(import_status, dict):
        return None
    import_status_keys = set(import_status.keys())
    if not all(isinstance(key, str) for key in import_status_keys):
        return None
    for key, value_type in [
        ('success', bool),
        ('notes', list),
        ('utc_datetime', str),
        ('object_id', (int, type(None)))
    ]:
        if key not in import_status_keys:
            return None
        if not isinstance(import_status[key], value_type):  # type: ignore
            return None
    success = import_status['success']
    notes = import_status['notes']
    if not all(isinstance(note, str) for note in notes):
        return None
    utc_datetime = import_status['utc_datetime']
    try:
        datetime.datetime.strptime(utc_datetime, '%Y-%m-%d %H:%M:%S')
    except Exception:
        return None
    object_id = import_status['object_id']
    if success:
        if object_id is None or object_id <= 0:
            return None
    else:
        if object_id is not None:
            return None
    return ObjectShareImportStatus(
        success=success,
        notes=notes,
        utc_datetime=utc_datetime,
        object_id=object_id
    )



# coding: utf-8
"""

"""

from . import authentication
from . import automatic_schema_updates
from . import background_tasks
from . import dataverse_export
from . import default_permissions
from . import eln_imports
from . import favorites
from . import files
from . import file_log
from . import groups
from . import group_categories
from . import info_pages
from . import instruments
from . import instrument_log_entries
from . import instrument_translation
from . import locations
from . import location_log
from . import location_permissions
from . import markdown_to_html_cache
from . import markdown_images
from . import objects
from . import object_permissions
from . import projects
from . import scicat_export
from . import settings
from . import temporary_files
from . import topics
from . import users
from . import webhooks

from .actions import Action, ActionType, SciCatExportType
from .action_translations import ActionTranslation, ActionTypeTranslation
from .action_permissions import UserActionPermissions, GroupActionPermissions, ProjectActionPermissions, AllUserActionPermissions
from .api_log import APILogEntry, HTTPMethod
from .authentication import Authentication, AuthenticationType, TwoFactorAuthenticationMethod, SAMLArtifacts, SAMLMetadata, SAMLMetadataType
from .automatic_schema_updates import UpdatableObjectsCheck, UpdatableObjectsCheckStatus, AutomaticSchemaUpdate, AutomaticSchemaUpdateStatus
from .background_tasks import BackgroundTask, BackgroundTaskStatus
from .comments import Comment
from .components import Component
from .component_authentication import ComponentAuthentication, OwnComponentAuthentication, ComponentAuthenticationType
from .dataverse_export import DataverseExport, DataverseExportStatus
from .default_permissions import DefaultUserPermissions, DefaultGroupPermissions, DefaultProjectPermissions, AllUserDefaultPermissions
from .download_service import DownloadServiceJobFile
from .eln_imports import ELNImport, ELNImportObject, ELNImportAction
from .favorites import FavoriteAction, FavoriteInstrument
from .fed_logs import FedUserLogEntry, FedUserLogEntryType, FedObjectLogEntry, FedObjectLogEntryType, FedLocationLogEntryType, FedLocationLogEntry, FedActionLogEntryType, FedActionLogEntry, FedActionTypeLogEntry, FedActionTypeLogEntryType, FedInstrumentLogEntry, FedInstrumentLogEntryType, FedCommentLogEntry, FedCommentLogEntryType, FedFileLogEntry, FedFileLogEntryType, FedObjectLocationAssignmentLogEntry, FedObjectLocationAssignmentLogEntryType, FedLocationTypeLogEntry, FedLocationTypeLogEntryType
from .files import File
from .file_log import FileLogEntry, FileLogEntryType
from .groups import Group
from .group_categories import GroupCategory
from .info_pages import InfoPage, InfoPageAcknowledgement
from .instruments import Instrument
from .instrument_log_entries import InstrumentLogEntry
from .instrument_translation import InstrumentTranslation
from .languages import Language
from .locations import Location, ObjectLocationAssignment, LocationType, LocationCapacity
from .location_log import LocationLogEntry, LocationLogEntryType
from .location_permissions import AllUserLocationPermissions, UserLocationPermissions, GroupLocationPermissions, ProjectLocationPermissions
from .markdown_to_html_cache import MarkdownToHTMLCacheEntry
from .markdown_images import MarkdownImage
from .minisign_keys import KeyPair
from .notifications import Notification, NotificationType, NotificationMode, NotificationModeForType
from .objects import Objects, Object
from .object_log import ObjectLogEntry, ObjectLogEntryType
from .object_permissions import UserObjectPermissions, GroupObjectPermissions, ProjectObjectPermissions, AllUserObjectPermissions, AnonymousUserObjectPermissions
from .object_publications import ObjectPublication
from .permissions import Permissions
from .projects import Project, UserProjectPermissions, GroupProjectPermissions, SubprojectRelationship
from .scicat_export import SciCatExport
from .settings import Settings
from .shares import ObjectShare
from .tags import Tag
from .topics import Topic
from .temporary_files import TemporaryFile
from .users import User, UserType, UserFederationAlias, FederatedIdentity
from .user_log import UserLogEntry, UserLogEntryType
from .webhooks import Webhook, WebhookType


__all__ = [
    'api_log',
    'authentication',
    'automatic_schema_updates',
    'background_tasks',
    'dataverse_export',
    'default_permissions',
    'eln_imports',
    'favorites',
    'files',
    'file_log',
    'groups',
    'group_categories',
    'info_pages',
    'instruments',
    'instrument_log_entries',
    'instrument_translation',
    'locations',
    'location_log',
    'location_permissions',
    'markdown_to_html_cache',
    'markdown_images',
    'objects',
    'object_permissions',
    'projects',
    'scicat_export',
    'settings',
    'temporary_files',
    'topics',
    'users',
    'webhooks',
    'Action',
    'ActionType',
    'ActionTranslation',
    'ActionTypeTranslation',
    'AutomaticSchemaUpdate',
    'AutomaticSchemaUpdateStatus',
    'BackgroundTask',
    'BackgroundTaskStatus',
    'UserActionPermissions',
    'GroupActionPermissions',
    'ProjectActionPermissions',
    'AllUserActionPermissions',
    'APILogEntry',
    'Authentication',
    'AuthenticationType',
    'Comment',
    'DataverseExport',
    'DataverseExportStatus',
    'DownloadServiceJobFile',
    'ELNImport',
    'ELNImportObject',
    'ELNImportAction',
    'FavoriteAction',
    'FavoriteInstrument',
    'File',
    'FileLogEntry',
    'FileLogEntryType',
    'Group',
    'GroupCategory',
    'HTTPMethod',
    'InfoPage',
    'InfoPageAcknowledgement',
    'Instrument',
    'InstrumentTranslation',
    'InstrumentLogEntry',
    'Language',
    'Location',
    'LocationCapacity',
    'LocationType',
    'LocationLogEntry',
    'LocationLogEntryType',
    'AllUserLocationPermissions',
    'UserLocationPermissions',
    'GroupLocationPermissions',
    'ProjectLocationPermissions',
    'ObjectLocationAssignment',
    'MarkdownToHTMLCacheEntry',
    'MarkdownImage',
    'Notification',
    'NotificationType',
    'NotificationMode',
    'NotificationModeForType',
    'Objects',
    'Object',
    'ObjectLogEntry',
    'ObjectLogEntryType',
    'ObjectPublication',
    'UserObjectPermissions',
    'GroupObjectPermissions',
    'ProjectObjectPermissions',
    'AllUserObjectPermissions',
    'AnonymousUserObjectPermissions',
    'DefaultUserPermissions',
    'DefaultGroupPermissions',
    'DefaultProjectPermissions',
    'AllUserDefaultPermissions',
    'Permissions',
    'Project',
    'UserProjectPermissions',
    'GroupProjectPermissions',
    'SciCatExport',
    'SciCatExportType',
    'SubprojectRelationship',
    'Settings',
    'Tag',
    'TemporaryFile',
    'Topic',
    'TwoFactorAuthenticationMethod',
    'SAMLArtifacts',
    'SAMLMetadata',
    'SAMLMetadataType',
    'UpdatableObjectsCheck',
    'UpdatableObjectsCheckStatus',
    'User',
    'UserFederationAlias',
    'FederatedIdentity',
    'UserType',
    'UserLogEntry',
    'UserLogEntryType',
    'Component',
    'ComponentAuthentication',
    'OwnComponentAuthentication',
    'ComponentAuthenticationType',
    'ObjectShare',
    'FedUserLogEntry',
    'FedUserLogEntryType',
    'FedObjectLogEntry',
    'FedObjectLogEntryType',
    'FedLocationLogEntry',
    'FedLocationLogEntryType',
    'FedLocationTypeLogEntry',
    'FedLocationTypeLogEntryType',
    'FedActionLogEntry',
    'FedActionLogEntryType',
    'FedActionTypeLogEntry',
    'FedActionTypeLogEntryType',
    'FedInstrumentLogEntry',
    'FedInstrumentLogEntryType',
    'FedCommentLogEntry',
    'FedCommentLogEntryType',
    'FedFileLogEntry',
    'FedFileLogEntryType',
    'FedObjectLocationAssignmentLogEntry',
    'FedObjectLocationAssignmentLogEntryType',
    'Webhook',
    'WebhookType',
    'KeyPair'
]



"""
Wrapper module for Flask-MonitoringDashboard

This module creates an extension-like interface for Flask-MonitoringDashboard
and ensures that nothing is imported from that package if it should not be
used, as the package already performs some logic on import.

Also, this module defines replacement functions for the authentication process
of the dashboard, so that the existing Flask-Login based authentication system
is used.
"""
import functools
import os
import secrets
import typing

import flask
import flask_login
import werkzeug.exceptions

from . import version

F = typing.TypeVar('F', bound=typing.Callable[..., typing.Any])


def _secure(func: F) -> F:
    @flask_login.login_required
    @functools.wraps(func)
    def wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        if not _is_admin():
            return flask.abort(403)
        return func(*args, **kwargs)
    return typing.cast(F, wrapper)


def _is_admin() -> bool:
    return bool(flask_login.current_user.is_authenticated and flask_login.current_user.is_admin)


def _on_login(user: typing.Any) -> None:
    if not _is_admin():
        # login via dashboard is disabled
        raise werkzeug.exceptions.Forbidden()


def _on_logout() -> werkzeug.Response:
    flask_login.logout_user()
    return flask.redirect(flask.url_for('frontend.index'))


def _get_current_user_id() -> typing.Optional[int]:
    if flask_login.current_user.is_authenticated:
        return flask_login.current_user.get_id()
    return None


def _get_ip() -> str:
    return flask.request.headers.get('X-Forwarded-For', flask.request.environ['REMOTE_ADDR'])


def init_app(app: flask.Flask) -> None:
    # late import to only create dashboard Blueprint if it should be used
    try:
        import flask_monitoringdashboard as dashboard
    except ImportError:
        # set TZ variable to prevent crash in tzlocal 2.0.0 get_localzone
        if not os.environ.get('TZ'):
            os.environ['TZ'] = 'UTC'
        # try importing again
        import flask_monitoringdashboard as dashboard

    # replace authentication functions to use flask_login instead
    from flask_monitoringdashboard.core import auth
    auth.secure = _secure
    auth.admin_secure = _secure
    auth.is_admin = _is_admin
    auth.on_login = _on_login
    auth.on_logout = _on_logout

    dashboard.config.link = 'admin/dashboard'
    dashboard.config.version = version.__version__
    dashboard.config.brand_name = 'SampleDB Monitoring Dashboard'
    dashboard.config.title_name = 'SampleDB Monitoring Dashboard'
    dashboard.config.description = 'Automatically monitor the evolving performance of SampleDB'
    dashboard.config.show_login_banner = False
    dashboard.config.show_login_footer = False
    dashboard.config.table_prefix = 'dashboard_'
    dashboard.config.security_token = secrets.token_hex(32)
    dashboard.config.password = secrets.token_hex(32)
    dashboard.config.group_by = _get_current_user_id
    dashboard.config.get_ip = _get_ip
    dashboard.config.database_name = app.config['MONITORINGDASHBOARD_DATABASE']
    dashboard.bind(app)
